/*--------------------------------------------------------------------
This source distribution is placed in the public domain by its author,
Jason Papadopoulos. You may use it for any purpose, free of charge,
without having to notify anyone. I disclaim any responsibility for any
errors.

Optionally, please be nice and tell me if you find this source to be
useful. Again optionally, if you add to the functionality present here
please consider making those additions public too, so that others may 
benefit from your work.	
       				   --jasonp@boo.net 1/5/07
--------------------------------------------------------------------*/

#include "filter.h"

/*--------------------------------------------------------------------*/
static void dump_cycles(msieve_obj *obj, merge_t *merge) {

	uint32 i;
	relation_set_t *relset_array = merge->relset_array;
	uint32 num_relsets = merge->num_relsets;
	char buf[256];
	FILE *cycle_fp;

	sprintf(buf, "%s.cyc", obj->savefile_name);
	cycle_fp = fopen(buf, "wb");
	if (cycle_fp == NULL) {
		logprintf(obj, "error: can't open cycle file\n");
		exit(-1);
	}

	fwrite(&num_relsets, (size_t)1, sizeof(uint32), cycle_fp);

	for (i = 0; i < num_relsets; i++) {
		relation_set_t *r = relset_array + i;
		uint32 num = r->num_relations;

		r->relation_list[num - 1] |= (uint32)(0x80000000);
		fwrite(r->relation_list, (size_t)num,
				sizeof(uint32), cycle_fp);
		free(r->relation_list);
	}
	fclose(cycle_fp);
	free(merge->relset_array);
	merge->relset_array = NULL;
}

/*--------------------------------------------------------------------*/
#define SECOND_PASS_FRACTION 0.5

uint32 nfs_filter_relations(msieve_obj *obj, 
		sieve_param_t *params, mp_t *n) {

	filter_t filter;
	merge_t merge;
	uint32 entries_r;
	uint32 entries_a;
	uint32 filtmin_r_pass2;
	uint32 filtmin_a_pass2;
	uint32 extra_needed;
	factor_base_t fb;

	logprintf(obj, "\n");
	logprintf(obj, "commencing relation filtering\n");

	/* generate or read the factor base. We don't need
	   the complete factor base to do the filtering, so
	   free the largest structures to save some memory */

	memset(&fb, 0, sizeof(fb));
	fb.rfb.max_prime = params->rfb_limit;
	fb.afb.max_prime = params->afb_limit;
	if (read_factor_base(obj, n, &fb)) {
		create_factor_base(obj, &fb, 1);
		write_factor_base(obj, n, &fb);
	}
	entries_r = (uint32)(SECOND_PASS_FRACTION * fb.rfb.num_entries);
	entries_a = (uint32)(SECOND_PASS_FRACTION * fb.afb.num_entries);
	filtmin_r_pass2 = fb.rfb.entries[entries_r].p;
	filtmin_a_pass2 = fb.afb.entries[entries_a].p;
	free(fb.rfb.entries);
	free(fb.afb.entries);
	fb.rfb.entries = NULL;
	fb.afb.entries = NULL;

	/* delete duplicate relations */

	nfs_purge_duplicates(obj, &fb);

	/* singleton removal happens in two phases. The first
	   phase does the singleton removal and clique processing
	   for all the ideals larger than the factor base limits. 
	   
	   We avoid pruning too many cliques, to leave some room
	   for the second phase */

	memset(&filter, 0, sizeof(filter));
	entries_r = fb.rfb.num_entries;
	entries_a = fb.afb.num_entries;
	filter.filtmin_r = fb.rfb.max_prime;
	filter.filtmin_a = fb.afb.max_prime;
	filter.final_excess = entries_r + entries_a;
	filter.initial_excess = (uint32)(2.8 * filter.final_excess);
	filter.merge_excess = (uint32)(1.7 * filter.final_excess);

	extra_needed = nfs_purge_singletons(obj, &fb, &filter, 1);
	if (extra_needed > 0)
		return extra_needed;

	/* throw away the list of relations that was generated */

	free(filter.relation_array);
	filter.relation_array = NULL;

	/* repeat the singleton removal and clique processing,
	   but now reach into the factor base for large
	   ideals. This means relations will have many more 
	   ideals in them, but the first phase removed so many
	   relations already that the memory use is manageable.
	   This two-step process is how the pros filter :) 
	   
	   Note that what happens below is a little tricky.
	   The expected dimension of the final matrix is about
	   90% of the size of the factor base, but we keep information
	   on all ideals in the last 50% of the factor base, as well
	   as ideals larger than the factor base bound. This technically 
	   means the merge phase can produce a matrix that's half the 
	   size we're shooting for, but that matrix would be hopelessly 
	   dense. Instead, the merge code will use the extra information
	   to skip merging ideals that appear in many relations. 
	   The ignored ideals increase the number of cycles that 
	   must be found, and we keep skipping ideals until we're required
	   to get the number of cycles corresponding to the matrix size
	   that's expected.

	   My experiments show that compared to just constructing the 
	   correct-size matrix, this technique causes a 30% reduction 
	   in the number of nonzero matrix entries! It also allows
	   much of the postprocessing typically applied to NFS matrices
	   (deletion of empty rows, combining cliques, etc.) to take 
	   place early, by code that is optimized to handle it */

	filter.filtmin_r = filtmin_r_pass2;
	filter.filtmin_a = filtmin_a_pass2;
	filter.final_excess = 0.9 * (entries_r + entries_a) +
				     NUM_EXTRA_RELATIONS;
	filter.initial_excess = filter.merge_excess = entries_r + entries_a +
							NUM_EXTRA_RELATIONS;

	extra_needed = nfs_purge_singletons(obj, &fb, &filter, 0);
	if (extra_needed > 0)
		return extra_needed;

	/* finally, merge all of the relations that are left */

	extra_needed = SECOND_PASS_FRACTION * (entries_r + entries_a);
	nfs_merge_init(obj, &filter);
	nfs_merge_2way(obj, &filter, &merge);
	nfs_merge_full(obj, &merge, extra_needed, filter.final_excess);
	dump_cycles(obj, &merge);
	return 0;
}
