/*--------------------------------------------------------------------
This source distribution is placed in the public domain by its author,
Jason Papadopoulos. You may use it for any purpose, free of charge,
without having to notify anyone. I disclaim any responsibility for any
errors.

Optionally, please be nice and tell me if you find this source to be
useful. Again optionally, if you add to the functionality present here
please consider making those additions public too, so that others may 
benefit from your work.	
       				   --jasonp@boo.net 1/5/07
--------------------------------------------------------------------*/

#include "filter.h"

typedef struct {
	list_t *heap_relsets;
	list_t *heap_ideals;
	list_t *hashtable;
	uint32 num_bins;
	uint32 next_bin;
	uint32 worst_bin;
} discrete_heap_t;

#define MAX_RELSETS 100

/*--------------------------------------------------------------------*/
static int compare_uint32(const void *x, const void *y) {
	uint32 *xx = (uint32 *)x;
	uint32 *yy = (uint32 *)y;
	if (*xx > *yy)
		return 1;
	if (*xx < *yy)
		return -1;
	return 0;
}

/*--------------------------------------------------------------------*/
static uint32 heap_compute_key(list_t *heap_ideal) {

	if (heap_ideal->payload.ideal_info.num_relsets == 0)
		return (uint32)(-1);

	return ((uint32)heap_ideal->payload.ideal_info.num_relsets - 1) *
	       ((uint32)heap_ideal->payload.ideal_info.min_relset_size - 1);
}

/*--------------------------------------------------------------------*/
static void heap_add_ideal(discrete_heap_t *heap, 
				uint32 ideal) {

	list_t *head;
	list_t *heap_ideal = heap->heap_ideals + ideal;
	uint32 key = heap_compute_key(heap_ideal);

	if (key == (uint32)(-1)) {
		printf("error: attempted to heapify an empty ideal\n");
		exit(-1);
	}
	if (key >= heap->num_bins) {
		uint32 i;
		uint32 new_size = MAX(key, 2 * heap->num_bins);
		list_t *new_hashtable = (list_t *)malloc(new_size *
							sizeof(list_t));
		for (i = 0; i < heap->num_bins; i++) {
			list_t *old_entry = heap->hashtable + i;
			list_t *new_entry = new_hashtable + i;
			new_entry->next = new_entry;
			new_entry->prev = new_entry;
			if (old_entry->next != old_entry) {
				uint32 j;

				j = old_entry->next - heap->heap_ideals;
				new_entry->next = old_entry->next;
				heap->heap_ideals[j].prev = new_entry;

				j = old_entry->prev - heap->heap_ideals;
				new_entry->prev = old_entry->prev;
				heap->heap_ideals[j].next = new_entry;
			}
		}
		for (; i < new_size; i++) {
			list_t *entry = new_hashtable + i;
			entry->next = entry;
			entry->prev = entry;
		}
		free(heap->hashtable);
		heap->hashtable = new_hashtable;
		heap->num_bins = new_size;
	}
	head = heap->hashtable + key;
	heap_ideal->prev = head;
	heap_ideal->next = head->next;
	head->next->prev = heap_ideal;
	head->next = heap_ideal;
	heap->next_bin = MIN(heap->next_bin, key);
	heap->worst_bin = MAX(heap->worst_bin, key);
}
	

/*--------------------------------------------------------------------*/
static void heap_remove_ideal(discrete_heap_t *heap, 
				uint32 ideal) {

	list_t *heap_ideal = heap->heap_ideals + ideal;
	uint32 key = heap_compute_key(heap_ideal);
	list_t *head;

	heap_ideal->next->prev = heap_ideal->prev;
	heap_ideal->prev->next = heap_ideal->next;
	heap_ideal->prev = heap_ideal;
	heap_ideal->next = heap_ideal;

	if (key == heap->next_bin) {
		head = heap->hashtable + key;
		while (key < heap->num_bins && head->next == head) {
			head++;
			key++;
		}
		heap->next_bin = key;
	}

	if (key == heap->worst_bin) {
		head = heap->hashtable + key;
		while ((int32)key >= 0  && head->next == head) {
			head--;
			key--;
		}
		heap->worst_bin = key;
	}
}
	

/*--------------------------------------------------------------------*/
static void heap_add_relset(discrete_heap_t *heap, 
				relation_set_t *relset,
				uint32 heapified) {
	
	uint32 i;
	list_t *ideal_list = relset->ideal_list;
	uint32 weight = relset->num_small_ideals +
			relset->num_large_ideals;

	for (i = 1; i <= relset->num_large_ideals; i++) {
		list_t *relset_ideal = ideal_list + i;
		uint32 ideal = relset_ideal->payload.ideal;
		list_t *heap_relset = heap->heap_relsets + ideal;
		list_t *heap_ideal = heap->heap_ideals + ideal;
		uint16 prev_weight = 
				heap_ideal->payload.ideal_info.min_relset_size;

		if (heapified)
			heap_remove_ideal(heap, ideal);

		heap_ideal->payload.ideal_info.num_relsets++;
		heap_ideal->payload.ideal_info.min_relset_size = 
						MIN(prev_weight, weight);
		relset_ideal->prev = heap_relset;
		relset_ideal->next = heap_relset->next;
		heap_relset->next->prev = relset_ideal;
		heap_relset->next = relset_ideal;

		if (heapified)
			heap_add_ideal(heap, ideal);
	}
}

/*--------------------------------------------------------------------*/
static void heap_remove_relset(discrete_heap_t *heap, 
				relation_set_t *relset,
				relation_set_t *relset_array) {
	
	uint32 i;
	list_t *ideal_list = relset->ideal_list;
	uint32 weight = relset->num_small_ideals +
			relset->num_large_ideals;

	for (i = 1; i <= relset->num_large_ideals; i++) {
		list_t *relset_ideal = ideal_list + i;
		uint32 ideal = relset_ideal->payload.ideal;
		list_t *heap_relset = heap->heap_relsets + ideal;
		list_t *heap_ideal = heap->heap_ideals + ideal;
		uint16 prev_weight = 
				heap_ideal->payload.ideal_info.min_relset_size;

		heap_remove_ideal(heap, ideal);

		relset_ideal->prev->next = relset_ideal->next;
		relset_ideal->next->prev = relset_ideal->prev;
		relset_ideal->next = relset_ideal;
		relset_ideal->prev = relset_ideal;

		if (--(heap_ideal->payload.ideal_info.num_relsets) == 0) {
			heap_ideal->payload.ideal_info.min_relset_size = 0xffff;
		}
		else {
			if (weight == prev_weight) {
				list_t *curr_relset = heap_relset->next;
				weight = 0xffff;
				while (curr_relset != heap_relset) {
					list_t *tmp = curr_relset;
					relation_set_t *r;
					while (tmp->next != NULL)
						tmp--;
					r = relset_array + 
						tmp->payload.relset_num;
					weight = MIN(weight, 
							r->num_small_ideals +
							r->num_large_ideals);
					curr_relset = curr_relset->next;
				}
				heap_ideal->payload.ideal_info.min_relset_size =
							weight;
			}

			heap_add_ideal(heap, ideal);
		}

	}
}

/*--------------------------------------------------------------------*/
static uint32 heap_remove_min(discrete_heap_t *heap) {

	uint32 ideal;
	uint32 key = heap->next_bin;

	if (key >= heap->num_bins)
		return (uint32)(-1);
	
	ideal = heap->hashtable[key].next - heap->heap_ideals;
	heap_remove_ideal(heap, ideal);
	return ideal;
}
	
/*--------------------------------------------------------------------*/
static uint32 heap_remove_worst(discrete_heap_t *heap) {

	uint32 ideal;
	uint32 key = heap->worst_bin;

	if (key == (uint32)(-1))
		return (uint32)(-1);
	
	ideal = heap->hashtable[key].next - heap->heap_ideals;
	heap_remove_ideal(heap, ideal);
	return ideal;
}
	
/*--------------------------------------------------------------------*/
static void heap_init(discrete_heap_t *heap, merge_t *merge) {

	uint32 i;
	relation_set_t *relset_array;
	uint32 num_relsets;
	uint32 num_ideals;

	relset_array = merge->relset_array;
	num_relsets = merge->num_relsets;
	num_ideals = merge->num_ideals;

	heap->num_bins = 5000;
	heap->next_bin = heap->num_bins;
	heap->worst_bin = 0;
	heap->hashtable = (list_t *)malloc(heap->num_bins * sizeof(list_t));
	heap->heap_ideals = (list_t *)malloc(num_ideals * sizeof(list_t));
	heap->heap_relsets = (list_t *)malloc(num_ideals * sizeof(list_t));

	for (i = 0; i < heap->num_bins; i++) {
		list_t *entry = heap->hashtable + i;
		entry->next = entry;
		entry->prev = entry;
	}
	for (i = 0; i < num_ideals; i++) {
		list_t *entry = heap->heap_relsets + i;
		entry->next = entry;
		entry->prev = entry;

		entry = heap->heap_ideals + i;
		entry->payload.ideal_info.num_relsets = 0;
		entry->payload.ideal_info.min_relset_size = 0xffff;
		entry->next = entry;
		entry->prev = entry;
	}
	for (i = 0; i < num_relsets; i++) {
		relation_set_t *relset = relset_array + i;
		list_t *ideal_list = relset->ideal_list;
	
		if (ideal_list)
			heap_add_relset(heap, relset, 0);
	}
	for (i = 0; i < num_ideals; i++) {
		heap_add_ideal(heap, i);
	}
}

/*--------------------------------------------------------------------*/
static void heap_free(discrete_heap_t *heap) {
	free(heap->hashtable);
	free(heap->heap_relsets);
	free(heap->heap_ideals);
}

/*--------------------------------------------------------------------*/
static int compare_relsets(const void *x, const void *y) {
	relation_set_t *xx = (relation_set_t *)x;
	relation_set_t *yy = (relation_set_t *)y;
	return (int)xx->num_small_ideals - (int)yy->num_small_ideals;
}

/*--------------------------------------------------------------------*/
static void do_merges_core(relation_set_t *relsets, 
				uint32 num_relsets) {
	
	uint32 weight;
	uint32 i;
	relation_set_t pivot, tmp_relset;
	uint32 pivot_idx;

	if (num_relsets == 1) {
		free(relsets[0].relation_list);
		free(relsets[0].ideal_list);
		return;
	}

	memset(&tmp_relset, 0, sizeof(tmp_relset));
	weight = relsets[0].num_small_ideals +
		 relsets[0].num_large_ideals;
	pivot = relsets[0];
	pivot_idx = 0;
	for (i = 1; i < num_relsets; i++) {
		uint32 new_weight = relsets[i].num_small_ideals +
					relsets[i].num_large_ideals;
		if (new_weight < weight) {
			pivot = relsets[i];
			pivot_idx = i;
			weight = new_weight;
		}
	}
	relsets[pivot_idx] = relsets[i-1];
		
	for (i = 0; i < num_relsets - 1; i++) {
		uint32 j, k, m;
		uint32 num_ideals1 = relsets[i].num_large_ideals;
		uint32 num_ideals2 = pivot.num_large_ideals;

		tmp_relset.num_relations = relsets[i].num_relations +
						pivot.num_relations;
		tmp_relset.num_small_ideals = relsets[i].num_small_ideals +
						pivot.num_small_ideals;
		tmp_relset.relation_list = (uint32 *)malloc(sizeof(uint32) *
						tmp_relset.num_relations);
		for (j = 0; j < relsets[i].num_relations; j++) {
			tmp_relset.relation_list[j] = 
					relsets[i].relation_list[j];
		}
		for (k = 0; k < pivot.num_relations; j++, k++) {
			tmp_relset.relation_list[j] = 
					pivot.relation_list[k];
		}
		if (num_ideals1 == 1 && num_ideals2 == 1) {
			tmp_relset.num_large_ideals = 0;
			tmp_relset.ideal_list = NULL;
			free(relsets[i].relation_list);
			free(relsets[i].ideal_list);
			relsets[i] = tmp_relset;
			continue;
		}

		tmp_relset.ideal_list = (list_t *)malloc(sizeof(list_t) *
						(num_ideals1+num_ideals2-2+1));
		j = k = 1;
		m = 0;
		while (j <= num_ideals1 && k <= num_ideals2) {
			uint32 ideal1 = relsets[i].ideal_list[j].payload.ideal;
			uint32 ideal2 = pivot.ideal_list[k].payload.ideal;
			if (ideal1 < ideal2) {
				tmp_relset.ideal_list[m+1].payload.ideal = 
								ideal1;
				j++; m++;
			}
			else if (ideal1 > ideal2) {
				tmp_relset.ideal_list[m+1].payload.ideal = 
								ideal2;
				k++; m++;
			}
			else {
				j++; k++;
			}
		}
		for (; j <= num_ideals1; j++, m++) {
			uint32 ideal1 = relsets[i].ideal_list[j].payload.ideal;
			tmp_relset.ideal_list[m+1].payload.ideal = ideal1;
		}
		for (; k <= num_ideals2; k++, m++) {
			uint32 ideal2 = pivot.ideal_list[k].payload.ideal;
			tmp_relset.ideal_list[m+1].payload.ideal = ideal2;
		}

		if (m == 0) {
			free(tmp_relset.ideal_list);
			tmp_relset.ideal_list = NULL;
		}
		tmp_relset.num_large_ideals = m;
		free(relsets[i].relation_list);
		free(relsets[i].ideal_list);
		relsets[i] = tmp_relset;
	}

	free(pivot.relation_list);
	free(pivot.ideal_list);
}

/*--------------------------------------------------------------------*/
static void bury_one_ideal(relation_set_t *relsets, 
				uint32 num_relsets, uint32 ideal) {
	
	uint32 i, j;

	for (i = 0; i < num_relsets; i++) {
		relation_set_t *r = relsets + i;
		uint32 large_ideals = r->num_large_ideals;

		r->num_small_ideals++;
		if (--(r->num_large_ideals) == 0) {
			free(r->ideal_list);
			r->ideal_list = NULL;
		}
		else {
			for (j = 1; j <= large_ideals; j++) {
				if (r->ideal_list[j].payload.ideal == ideal)
					break;
			}
			for (j++; j <= large_ideals; j++) {
				r->ideal_list[j-1] = r->ideal_list[j];
			}
		}
	}
}

/*--------------------------------------------------------------------*/
#define NUM_IDEAL_BINS 6

void nfs_merge_init(msieve_obj *obj, filter_t *filter) {

	uint32 i;
	relation_ideal_t *curr_relation = filter->relation_array;
	uint32 num_relations = filter->num_relations;
	uint32 ideal_count_bins[NUM_IDEAL_BINS+2] = {0};

	for (i = 0; i < num_relations; i++) {
		uint32 num_ideals = curr_relation->ideal_count;
		if (num_ideals > NUM_IDEAL_BINS)
			ideal_count_bins[NUM_IDEAL_BINS+1]++;
		else
			ideal_count_bins[num_ideals]++;

		if (num_ideals > 1) {
			qsort(curr_relation->ideal_list, (size_t)num_ideals,
					sizeof(uint32), compare_uint32);
		}
		curr_relation = next_relation_ptr(curr_relation);
	}

	for (i = 0; i < NUM_IDEAL_BINS+1; i++) {
		logprintf(obj, "relations with %u large ideals: %u\n", 
					i, ideal_count_bins[i]);
	}
	logprintf(obj, "relations with %u+ large ideals: %u\n", 
				i, ideal_count_bins[i]);
}

/*--------------------------------------------------------------------*/
#define MAX_2WAY_IDEALS 100

void nfs_merge_2way(msieve_obj *obj, filter_t *filter,
			merge_t *merge) {

	uint32 i, j;
	ideal_map_t *ideal_map;
	relation_ideal_t *relation_array;
	relation_ideal_t *curr_relation;
	uint32 num_relations;
	uint32 num_ideals;
	uint32 num_merged;
	uint32 relset_memalloc;

	ideal_relation_t *reverse_array;
	uint32 num_reverse;
	uint32 num_reverse_alloc;

	relation_set_t *relset_array;
	uint32 num_relset;
	uint32 num_relset_alloc;

	logprintf(obj, "commencing 2-way merge\n");

	relation_array = filter->relation_array;
	num_relations = filter->num_relations;
	num_ideals = filter->num_ideals;
	ideal_map = (ideal_map_t *)calloc((size_t)num_ideals, 
					sizeof(ideal_map_t));

	num_reverse = 1;
	num_reverse_alloc = 10000;
	reverse_array = (ideal_relation_t *)malloc(num_reverse_alloc *
					sizeof(ideal_relation_t));

	curr_relation = relation_array;
	for (i = 0; i < num_relations; i++) {
		curr_relation->connected = 0;
		for (j = 0; j < curr_relation->ideal_count; j++) {
			uint32 ideal = curr_relation->ideal_list[j];
			ideal_map[ideal].payload++;
		}
		curr_relation = next_relation_ptr(curr_relation);
	}

	for (i = 0; i < num_ideals; i++) {
		if (ideal_map[i].payload == 2) {
			ideal_map[i].payload = 0;
			ideal_map[i].clique = 1;
		}
	}

	curr_relation = relation_array;
	for (i = 0; i < num_relations; i++) {
		for (j = 0; j < curr_relation->ideal_count; j++) {
			uint32 ideal = curr_relation->ideal_list[j];

			if (!ideal_map[ideal].clique)
				continue;

			if (num_reverse == num_reverse_alloc) {
				num_reverse_alloc *= 2;
				reverse_array = (ideal_relation_t *)realloc(
						reverse_array,
						num_reverse_alloc *
						sizeof(ideal_relation_t));
			}
			reverse_array[num_reverse].relation_array_word =
					(uint32 *)curr_relation -
					(uint32 *)relation_array;
			reverse_array[num_reverse].next = 
						ideal_map[ideal].payload;
			ideal_map[ideal].payload = num_reverse++;
		}

		curr_relation = next_relation_ptr(curr_relation);
	}

	num_relset = 0;
	num_relset_alloc = 10000;
	relset_array = (relation_set_t *)malloc(num_relset_alloc *
					sizeof(relation_set_t));

	curr_relation = relation_array;
	num_merged = 0;
	relset_memalloc = 0;
	for (i = 0; i < num_relations; 
		i++, curr_relation = next_relation_ptr(curr_relation)) {

		uint32 tmp_relations[MAX_RELSETS];
		uint32 tmp_ideals[MAX_2WAY_IDEALS];
		uint32 accum_ideals[MAX_2WAY_IDEALS];
		uint32 num_tmp_relation = 0;
		uint32 num_tmp_ideal = 0;
		uint32 num_small_ideal;
		relation_set_t *curr_relset;

		if (curr_relation->connected)
			continue;

		tmp_relations[num_tmp_relation++] = curr_relation->rel_index;
		num_small_ideal = curr_relation->gf2_factors;
		num_tmp_ideal = curr_relation->ideal_count;
		for (j = 0; j < num_tmp_ideal; j++)
			tmp_ideals[j] = curr_relation->ideal_list[j];
		curr_relation->connected = 1;

		for (j = 0; j < num_tmp_ideal; j++) {
			uint32 ideal = tmp_ideals[j];
			uint32 offset;

			if (!ideal_map[ideal].clique)
				continue;

			offset = ideal_map[ideal].payload;
			while (offset) {

				ideal_relation_t *rev = reverse_array + offset;
				relation_ideal_t *r = (relation_ideal_t *)
						((uint32 *)relation_array +
						rev->relation_array_word);
				if (r->connected) {
					offset = rev->next;
					continue;
				}

				tmp_relations[num_tmp_relation++] = 
							r->rel_index;
				num_small_ideal += r->gf2_factors;
				num_tmp_ideal = merge_relations(accum_ideals,
						tmp_ideals, num_tmp_ideal,
						r->ideal_list, r->ideal_count);
				memcpy(tmp_ideals, accum_ideals, 
						num_tmp_ideal * sizeof(uint32));
				num_merged++;
				r->connected = 1;
				j = (uint32)(-1);
				offset = rev->next;
			}
		}

		if (num_tmp_ideal >= MAX_2WAY_IDEALS || 
		    num_tmp_relation >= MAX_RELSETS) {
			printf("error: 2-way merge resources exceeded\n");
			exit(-1);
		}
		if (num_relset == num_relset_alloc) {
			num_relset_alloc *= 2;
			relset_array = (relation_set_t *)realloc(relset_array,
							num_relset_alloc *
							sizeof(relation_set_t));
		}
		curr_relset = relset_array + num_relset;
		curr_relset->num_relations = num_tmp_relation;
		curr_relset->num_small_ideals = num_small_ideal;
		curr_relset->num_large_ideals = num_tmp_ideal;

		curr_relset->relation_list = (uint32 *)
				malloc(num_tmp_relation * sizeof(uint32));
		for (j = 0; j < num_tmp_relation; j++)
			curr_relset->relation_list[j] = tmp_relations[j];

		curr_relset->ideal_list = NULL;
		if (num_tmp_ideal) {
			list_t *ideal_list = (list_t *)malloc(
					(num_tmp_ideal + 1) * sizeof(list_t));
			ideal_list[0].payload.relset_num = num_relset;
			ideal_list[0].prev = NULL;
			ideal_list[0].next = NULL;
			for (j = 1; j <= num_tmp_ideal; j++) {
				ideal_list[j].payload.ideal = tmp_ideals[j-1];
			}
			curr_relset->ideal_list = ideal_list;
		}
		num_relset++;
		relset_memalloc += num_tmp_relation * sizeof(uint32) +
					(num_tmp_ideal + 1) * sizeof(list_t);
	}

	logprintf(obj, "memory use: %.1f MB\n", (double)
			((curr_relation - relation_array) * 
						sizeof(relation_ideal_t) +
			 sizeof(ideal_map_t) * num_ideals +
			 num_reverse_alloc * sizeof(ideal_relation_t) +
			 num_relset_alloc * sizeof(relation_set_t) +
			 relset_memalloc) / 1048576);
	logprintf(obj, "merged %u relations\n", num_merged);

	free(reverse_array);
	free(filter->relation_array);
	filter->relation_array = NULL;
	relset_array = (relation_set_t *)realloc(relset_array,
				num_relset * sizeof(relation_set_t));
	memset(ideal_map, 0, num_ideals * sizeof(ideal_map_t));

	for (i = 0; i < num_relset; i++) {
		list_t *list = relset_array[i].ideal_list;
		for (j = 1; j <= relset_array[i].num_large_ideals; j++) {
			uint32 ideal = list[j].payload.ideal;
			ideal_map[ideal].payload++;
		}
	}
	for (i = j = 0; i < num_ideals; i++) {
		if (ideal_map[i].payload) {
			ideal_map[i].payload = j++;
		}
	}
	num_ideals = j;
	for (i = 0; i < num_relset; i++) {
		list_t *list = relset_array[i].ideal_list;
		for (j = 1; j <= relset_array[i].num_large_ideals; j++) {
			uint32 ideal = list[j].payload.ideal;
			list[j].payload.ideal = ideal_map[ideal].payload;
		}
	}

	logprintf(obj, "reduce to %u relation sets and %u "
			"unique ideals\n", num_relset, num_ideals);
	merge->relset_array = relset_array;
	merge->num_relsets = num_relset;
	merge->num_ideals = num_ideals;
	free(ideal_map);
}

/*--------------------------------------------------------------------*/
#define NUM_CYCLE_BINS 9

void nfs_merge_full(msieve_obj *obj, merge_t *merge,
			uint32 min_cycles, uint32 final_cycles) {

	uint32 i, j;
	relation_set_t *relset_array;
	uint32 num_relsets;
	uint32 num_ideals;
	uint32 num_cycles;
	discrete_heap_t heap;
	uint32 cycle_bins[NUM_CYCLE_BINS + 2] = {0};
	uint32 padded_final_cycles;
	uint32 max_cycles;

	uint32 curr_relset_idx[MAX_RELSETS];
	relation_set_t curr_relsets[MAX_RELSETS];
	uint32 num_curr_relsets;

	logprintf(obj, "commencing full merge\n");

	relset_array = merge->relset_array;
	num_relsets = merge->num_relsets;
	num_ideals = merge->num_ideals;
	for (i = num_cycles = 0; i < num_relsets; i++) {
		if (relset_array[i].num_large_ideals == 0)
			num_cycles++;
	}

	heap_init(&heap, merge);
	padded_final_cycles = 1.06 * final_cycles;
	
	while (1) {
		list_t *relset_head;
		list_t *r;
		uint32 ideal;

		if (++min_cycles >= final_cycles)
			break;

		ideal = heap_remove_worst(&heap);
		if (ideal == (uint32)(-1))
			break;

		num_curr_relsets = heap.heap_ideals[
					ideal].payload.ideal_info.num_relsets;
		if (num_curr_relsets >= MAX_RELSETS) {
			printf("error: too many relation sets\n");
			exit(-1);
		}

		relset_head = heap.heap_relsets + ideal;
		r = relset_head->next;
		num_curr_relsets = 0;

		while (r != relset_head) {
			list_t *tmp = r;

			while (tmp->next != NULL)
				tmp--;

			curr_relset_idx[num_curr_relsets] = 
					tmp->payload.relset_num;
			curr_relsets[num_curr_relsets] = 
					relset_array[tmp->payload.relset_num];
			num_curr_relsets++;
			r = r->next;
		}

		for (i = 0; i < num_curr_relsets; i++) {
			relation_set_t *r = relset_array + curr_relset_idx[i];
			heap_remove_relset(&heap, r, relset_array);
			r->relation_list = NULL;
			r->ideal_list = NULL;
		}

		bury_one_ideal(curr_relsets, num_curr_relsets, ideal);

		for (i = 0; i < num_curr_relsets; i++) {
			uint32 relset_num = curr_relset_idx[i];
			relation_set_t *r = relset_array + relset_num;
			*r = curr_relsets[i];
			if (r->num_large_ideals > 0) {
				r->ideal_list[0].next = NULL;
				r->ideal_list[0].prev = NULL;
				r->ideal_list[0].payload.relset_num = 
							relset_num;
				heap_add_relset(&heap, r, 1);
			}
			else {
				num_cycles++;
			}
		}
	}

	while (1) {
		list_t *relset_head;
		list_t *r;
		uint32 ideal;

		ideal = heap_remove_min(&heap);
		if (ideal == (uint32)(-1))
			break;

		num_curr_relsets = heap.heap_ideals[
					ideal].payload.ideal_info.num_relsets;
		if (num_curr_relsets >= MAX_RELSETS) {
			printf("error: too many relation sets\n");
			exit(-1);
		}

		relset_head = heap.heap_relsets + ideal;
		r = relset_head->next;
		num_curr_relsets = 0;

		while (r != relset_head) {
			list_t *tmp = r;

			while (tmp->next != NULL)
				tmp--;

			curr_relset_idx[num_curr_relsets] = 
					tmp->payload.relset_num;
			curr_relsets[num_curr_relsets] = 
					relset_array[tmp->payload.relset_num];
			num_curr_relsets++;
			r = r->next;
		}

		for (i = 0; i < num_curr_relsets; i++) {
			relation_set_t *r = relset_array + curr_relset_idx[i];
			heap_remove_relset(&heap, r, relset_array);
			r->relation_list = NULL;
			r->ideal_list = NULL;
		}

		do_merges_core(curr_relsets, num_curr_relsets);

		for (i = 0; i < num_curr_relsets - 1; i++) {
			uint32 relset_num = curr_relset_idx[i];
			relation_set_t *r = relset_array + relset_num;
			*r = curr_relsets[i];
			if (r->num_large_ideals > 0) {
				r->ideal_list[0].next = NULL;
				r->ideal_list[0].prev = NULL;
				r->ideal_list[0].payload.relset_num = 
							relset_num;
				heap_add_relset(&heap, r, 1);
			}
			else {
				num_cycles++;
			}
		}

		if (num_cycles >= padded_final_cycles)
			break;
	}

	for (i = num_cycles = 0; i < num_relsets; i++) {
		if (relset_array[i].relation_list &&
		    relset_array[i].num_large_ideals == 0) {
			relset_array[num_cycles++] = relset_array[i];
		}
		else {
			free(relset_array[i].relation_list);
			free(relset_array[i].ideal_list);
			relset_array[i].relation_list = NULL;
			relset_array[i].ideal_list = NULL;
		}
	}
	logprintf(obj, "found %u cycles, need %u\n", 
				num_cycles, final_cycles);

	qsort(relset_array, (size_t)num_cycles, 
			sizeof(relation_set_t), compare_relsets);

	num_cycles = MIN(final_cycles, num_cycles);
	max_cycles = 0;
	for (i = j = 0; i < num_cycles; i++) {
		j += relset_array[i].num_small_ideals;
		max_cycles = MAX(max_cycles, relset_array[i].num_relations);

		if (relset_array[i].num_relations > NUM_CYCLE_BINS)
			cycle_bins[NUM_CYCLE_BINS+1]++;
		else
			cycle_bins[relset_array[i].num_relations]++;
	}

	logprintf(obj, "weight of first %u cycles is about %u (%.2f/cycle)\n",
			num_cycles, j, (double)j / num_cycles);
	logprintf(obj, "distribution of cycle lengths:\n");
	for (i = 1; i < NUM_CYCLE_BINS + 1; i++)
		logprintf(obj, "%u relations: %u\n", i, cycle_bins[i]);
	logprintf(obj, "%u+ relations: %u\n", i, cycle_bins[i]);
	logprintf(obj, "heaviest cycle: %u relations\n", max_cycles);

	merge->num_relsets = num_cycles;
	heap_free(&heap);
}
