/*--------------------------------------------------------------------
This source distribution is placed in the public domain by its author,
Jason Papadopoulos. You may use it for any purpose, free of charge,
without having to notify anyone. I disclaim any responsibility for any
errors.

Optionally, please be nice and tell me if you find this source to be
useful. Again optionally, if you add to the functionality present here
please consider making those additions public too, so that others may 
benefit from your work.	
       				   --jasonp@boo.net 1/5/07
--------------------------------------------------------------------*/

#include <common.h>
#include "gnfs.h"

/* the number of quadratic characters for each
   matrix column. Bernstein writes that a QCB size
   of 50 'should be enough for any possible NFS
   factorization', but I originally used 48 here
   and it wasn't enough to generate an algebraic 
   square (for an input of only 100 digits!). The 
   matrix-building code is flexible enough so that 
   this number may be increased and everything 
   else will just work */

#define QCB_SIZE 65

/* the number of dense rows that are handled in a separate
   Gauss elimination phase after the Lanczos iteration
   completes. This is 64 - (the number of output dependencies 
   desired). The code should still work if this is set to
   zero, but I don't know why you would want to do that */

#define POST_LANCZOS_ROWS 48

/*--------------------------------------------------------------------*/
static int compare_uint32(const void *x, const void *y) {
	uint32 *xx = (uint32 *)x;
	uint32 *yy = (uint32 *)y;
	if (*xx > *yy)
		return 1;
	if (*xx < *yy)
		return -1;
	return 0;
}

/*--------------------------------------------------------------------*/
static int compare_weight(const void *x, const void *y) {
	la_col_t *xx = (la_col_t *)x;
	la_col_t *yy = (la_col_t *)y;
	return xx->weight - yy->weight;
}

/*------------------------------------------------------------------*/
static int compare_ideals(const void *x, const void *y) {
	
	/* used to determine the ordering of two ideals.
	   Ordering is by rational or algebraic type, then
	   by prime, then by root of prime (for algebraic
	   ideals). The ordering by prime is tricky, because 
	   -1 has a special value that must be explicitly 
	   accounted for */

	ideal_t *k = (ideal_t *)x;
	ideal_t *t = (ideal_t *)y;

	if (k->i.rat_or_alg == RATIONAL_IDEAL &&
	    t->i.rat_or_alg == ALGEBRAIC_IDEAL)
		return -1;
	if (k->i.rat_or_alg == ALGEBRAIC_IDEAL &&
	    t->i.rat_or_alg == RATIONAL_IDEAL)
		return 1;
		
	if (k->i.compressed_p == 0x7fffffff ||
	    t->i.compressed_p == 0x7fffffff) {
		if (k->i.compressed_p == 0x7fffffff &&
		    t->i.compressed_p != 0x7fffffff)
			return -1;
		else if (k->i.compressed_p != 0x7fffffff &&
		         t->i.compressed_p == 0x7fffffff)
			return 1;
		else
			return 0;
	}

	if (k->i.compressed_p < t->i.compressed_p)
		return -1;
	if (k->i.compressed_p > t->i.compressed_p)
		return 1;
	if (k->i.rat_or_alg == RATIONAL_IDEAL)
		return 0;
		
	if (k->i.r < t->i.r)
		return -1;
	if (k->i.r > t->i.r)
		return 1;
		
	return 0;
}

/*------------------------------------------------------------------*/
#define MAX_COL_IDEALS 1000

static uint32 combine_relations(la_col_t *col, relation_t *rlist,
				ideal_t *merged_ideals,
				fb_entry_t *qcb, uint32 *qcb_bits) {

	uint32 i, j, k;
	uint32 num_merged = 0;
	ideal_t tmp_ideals[MAX_COL_IDEALS];
	uint32 num_tmp_ideals;

	/* form the matrix column corresponding to a 
	   collection of relations */

	for (i = 0; i < col->cycle.num_relations; i++) {
		relation_t *r = rlist + col->cycle.list[i];
		relation_lp_t new_ideals;
		int64 a = (int64)(r->a);
		uint32 b = r->b;

		/* compute the quadratic characters for r */

		for (j = 0; j < QCB_SIZE; j++) {
			uint32 p = qcb[j].p;
			uint32 r = qcb[j].r;
			int64 res = a % (int64)p;
			int32 symbol;

			if (res < 0)
				res += (int64)p;

			symbol = mp_legendre_1(mp_modsub_1((uint32)res,
					mp_modmul_1(b, r, p), p), p);

			/* symbol must be 1 or -1; if it's 0,
			   there's something wrong with the choice
			   of primes in the QCB but this isn't
			   a fatal error */

			if (symbol == -1)
				qcb_bits[j / 32] ^= (uint32)1 << (j % 32);
			else if (symbol == 0)
				printf("warning: zero character\n");
		}

		/* get the ideal decomposition of relation i */

		if (find_large_ideals(r, &new_ideals, 0, 0) > MAX_IDEALS) {
			printf("error: overflow reading ideals\n");
			exit(-1);
		}

		/* merge it with the current list of ideals */

		j = k = 0;
		num_tmp_ideals = 0;
		while (j < num_merged && k < new_ideals.ideal_count) {
			int32 compare_result = compare_ideals(
						merged_ideals + j,
						new_ideals.ideal_list + k);
			if (compare_result < 0) {
				tmp_ideals[num_tmp_ideals++] = 
						merged_ideals[j++];
			}
			else if (compare_result > 0) {
				tmp_ideals[num_tmp_ideals++] = 
						new_ideals.ideal_list[k++];
			}
			else {
				j++; k++;
			}
		}
		while (j < num_merged) {
			tmp_ideals[num_tmp_ideals++] = merged_ideals[j++];
		}
		while (k < new_ideals.ideal_count) {
			tmp_ideals[num_tmp_ideals++] = 
						new_ideals.ideal_list[k++];
		}
		if (num_tmp_ideals >= MAX_COL_IDEALS) {
			printf("error: overflow merging ideals\n");
			exit(-1);
		}

		num_merged = num_tmp_ideals;
		memcpy(merged_ideals, tmp_ideals, 
				num_merged * sizeof(ideal_t));

		/* within the cycle structure, change from a pointer 
		   to relation i back to the relation number for i */

		col->cycle.list[i] = r->rel_index;

		/* if relation i will never be referenced again, free it */

		if (--(r->refcnt) == 0) {
			free(r->factors_r);
			free(r->factors_a);
		}
	}

	return num_merged;
}

/*------------------------------------------------------------------*/
static uint32 fill_small_ideals(ideal_t *small_ideals, 
				factor_base_t *fb) {
	uint32 i, j;
	uint32 last_p;
	fb_entry_t *entries;

	if (MAX_PACKED_PRIME == 0)
		return 0;

	/* fill in the rational ideal of -1 */

	small_ideals[0].i.rat_or_alg = RATIONAL_IDEAL;
	small_ideals[0].i.compressed_p = 0x7fffffff;
	small_ideals[0].i.r = (uint32)(-1);

	/* fill in the other small rational ideals. Note that
	   we only store a given prime once */

	entries = fb->rfb.entries;
	last_p = 0;
	for (i = 1, j = 0; entries[j].p <= MAX_PACKED_PRIME; j++) {

		uint32 p = entries[j].p;

		if (p == last_p)
			continue;

		small_ideals[i].i.rat_or_alg = RATIONAL_IDEAL;
		small_ideals[i].i.compressed_p = (p - 1) / 2;
		small_ideals[i].i.r = p;
		last_p = p;
		i++;
	}

	/* repeat for the algebraic ideals */

	entries = fb->afb.entries;
	for (j = 0; entries[j].p <= MAX_PACKED_PRIME; i++, j++) {
		small_ideals[i].i.rat_or_alg = ALGEBRAIC_IDEAL;
		small_ideals[i].i.compressed_p = (entries[j].p - 1) / 2;
		small_ideals[i].i.r = entries[j].r;
	}

	/* the roots of the ideals may not be in ascending order */

	qsort(small_ideals, (size_t)i, sizeof(ideal_t), compare_ideals);

	return i;
}

/*------------------------------------------------------------------*/
static void fill_qcb(fb_entry_t *qcb, factor_base_t *fb,
			uint32 min_qcb_ideal) {
	uint32 i, j;
	mp_poly_t *poly = &fb->afb.poly;
	prime_sieve_t sieve;

	/* construct the quadratic character base, starting
	   with the primes above min_qcb_ideal */

	init_prime_sieve(&sieve, min_qcb_ideal, 0xffffffff);

	i = 0;
	while (i < QCB_SIZE) {
		uint32 roots[MAX_POLY_DEGREE];
		uint32 num_roots, high_coeff;
		uint32 p = get_next_prime(&sieve);

		num_roots = poly_get_zeros(roots, poly, p, 
					&high_coeff, 0);

		/* p cannot be a projective root of the
		   algebraic poly */

		if (high_coeff == 0)
			continue;

		/* save the next batch of roots */

		for (j = 0; i < QCB_SIZE && j < num_roots; i++, j++) {
			qcb[i].p = p;
			qcb[i].r = roots[j];
		}
	}

	free_prime_sieve(&sieve);
}

/*------------------------------------------------------------------*/
#define LOG2_IDEAL_HASHTABLE_SIZE 20
#define MAX_SMALL_IDEALS 100

static uint32 build_matrix(msieve_obj *obj, factor_base_t *fb,
			uint32 *nrows, uint32 *dense_rows_out,
			uint32 *ncols, la_col_t **cols_out) {

	/* read in the relations that contribute to the
	   initial matrix, and form the quadratic characters
	   for each column */

	uint32 i, j, k;
	uint32 total_weight;
	uint32 num_relations;
	uint32 num_cycles;
	la_col_t *cycle_list;
	relation_t *rlist;
	hashtable_t unique_ideals;
	uint32 min_qcb_ideal;
	fb_entry_t qcb[QCB_SIZE];
	ideal_t small_ideals[MAX_SMALL_IDEALS];
	uint32 num_small_ideals;
	uint32 small_ideal_words;

	/* fill in the small ideals */

	num_small_ideals = fill_small_ideals(small_ideals, fb);

	/* we need an extra matrix row to make sure that dependencies 
	   involve an even number of relations. If the rational 
	   poly R(x) is monic, and we weren't using free relations,
	   the sign of R(x) is negative for all relations, meaning we 
	   already get the effect of this extra row. However, in 
	   general we can't assume both of these are true */
	
	small_ideal_words = (QCB_SIZE + 1 + num_small_ideals + 31) / 32;

	/* trim the factor base size */

	free(fb->rfb.entries);
	free(fb->afb.entries);
	fb->rfb.entries = NULL;
	fb->afb.entries = NULL;

	/* read in the cycles that form the matrix columns,
	   and the relations they will need */

	nfs_read_cycles(obj, fb, ncols, &cycle_list, 
			&num_relations, &rlist, 1, 0);
	num_cycles = *ncols;

	/* sort the list of factors of each relation, and find
	   the largest algebraic factor across the whole set */

	for (i = min_qcb_ideal = 0; i < num_relations; i++) {
		relation_t *r = rlist + i;
		qsort(r->factors_r, (size_t)r->num_factors_r, 
			sizeof(uint32), compare_uint32);
		qsort(r->factors_a, (size_t)r->num_factors_a, 
			sizeof(uint32), compare_uint32);
		min_qcb_ideal = MAX(min_qcb_ideal, 
				    r->factors_a[r->num_factors_a - 1]);
	}

	/* choose the quadratic character base from the primes
	   larger than any that appear in the list of relations */

	fill_qcb(qcb, fb, min_qcb_ideal + 1);

	/* now form the actual matrix columns */

	hashtable_init(&unique_ideals, LOG2_IDEAL_HASHTABLE_SIZE,
			10000, (uint32)(sizeof(ideal_t)/sizeof(uint32)));

	/* for each cycle */

	for (i = total_weight = 0; i < num_cycles; i++) {
		la_col_t *c = cycle_list + i;
		uint32 dense_rows[(QCB_SIZE+MAX_SMALL_IDEALS+31) / 32];
		ideal_t merged_ideals[MAX_COL_IDEALS];
		uint32 *mapped_ideals = (uint32 *)merged_ideals;
		uint32 num_merged;

		/* dense rows start off empty */

		for (j = 0; j < small_ideal_words; j++)
			dense_rows[j] = 0;

		/* merge the relations in the cycle and compute
		   the quadratic characters */

		num_merged = combine_relations(c, rlist, merged_ideals, 
						qcb, dense_rows);
		total_weight += num_merged;

		/* fill in the parity row, and place at 
		   bit position QCB_SIZE */

		if (c->cycle.num_relations % 2) {
			total_weight++;
			dense_rows[QCB_SIZE / 32] |= (uint32)1 << 
							(QCB_SIZE % 32);
		}

		/* assign a unique number to each ideal in 
		   the cycle. This will automatically ignore
		   empty rows in the matrix */

		for (j = k = 0; j < num_merged; j++) {
			ideal_t *ideal = merged_ideals + j;
			uint32 p = ideal->i.compressed_p;

			if (MAX_PACKED_PRIME > 0 && (p == 0x7fffffff || 
			    p <= (MAX_PACKED_PRIME-1)/2) ) {
				/* small ideal; store in compressed format */
				ideal_t *loc = (ideal_t *)bsearch(ideal, 
						small_ideals,
						(size_t)num_small_ideals,
						sizeof(ideal_t),
						compare_ideals);
				uint32 idx = QCB_SIZE + 1 +
						(loc - small_ideals);
				dense_rows[idx / 32] |= (uint32)1 << (idx % 32);
			}
			else {
				hash_t *entry = hashtable_find(&unique_ideals, 
							ideal->blob, NULL);
				mapped_ideals[k++] = hashtable_get_offset(
						&unique_ideals, entry);
			}
		}
		num_merged = k;

		/* save the concatenation of the sparse rows and
		   the dense rows */

		c->weight = num_merged;
		c->data = (uint32 *)malloc((num_merged + small_ideal_words) * 
						sizeof(uint32));
		memcpy(c->data, mapped_ideals, 
				num_merged * sizeof(uint32));
		for (j = 0; j < small_ideal_words; j++)
			c->data[num_merged + j] = dense_rows[j];
	}

	/* by this point, the only major structures in memory
	   should be the matrix columns; the relations themselves
	   were freed after they contributed their ideals to 
	   the columns.
	   
	   Note that at this point the number of rows does *not*
	   include the rows corresponding to the dense part of
	   the matrix */

	*cols_out = cycle_list;
	*nrows = hashtable_getall(&unique_ideals, NULL);
	*dense_rows_out = QCB_SIZE + 1 + num_small_ideals;
	hashtable_free(&unique_ideals);
	free(rlist);
	return total_weight;
}

/*--------------------------------------------------------------------*/
static void dump_cycles(msieve_obj *obj, la_col_t *cols, uint32 ncols) {

	uint32 i;
	char buf[256];
	FILE *cycle_fp;

	/* Note that the cycles are deleted after being dumped */

	sprintf(buf, "%s.cyc", obj->savefile_name);
	cycle_fp = fopen(buf, "wb");
	if (cycle_fp == NULL) {
		logprintf(obj, "error: can't open cycle file\n");
		exit(-1);
	}

	fwrite(&ncols, (size_t)1, sizeof(uint32), cycle_fp);

	for (i = 0; i < ncols; i++) {
		la_col_t *c = cols + i;
		uint32 num = c->cycle.num_relations;
		
		c->cycle.list[num - 1] |= (uint32)(0x80000000);
		fwrite(c->cycle.list, (size_t)num, sizeof(uint32), cycle_fp);
		free(c->cycle.list);
	}
	fclose(cycle_fp);
}

/*--------------------------------------------------------------------*/
static void dump_dependencies(msieve_obj *obj, 
			uint64 *deps, uint32 ncols) {

	char buf[256];
	FILE *deps_fp;

	/* we allow up to 64 dependencies, even though the
	   average case will have (64 - POST_LANCZOS_ROWS) */

	sprintf(buf, "%s.dep", obj->savefile_name);
	deps_fp = fopen(buf, "wb");
	if (deps_fp == NULL) {
		logprintf(obj, "error: can't open deps file\n");
		exit(-1);
	}

	fwrite(deps, (size_t)ncols, sizeof(uint64), deps_fp);
	fclose(deps_fp);
}

/*------------------------------------------------------------------*/
typedef struct {
	uint32 count;
	uint32 row_index;
} row_count_t;

static int compare_row_count(const void *x, const void *y) {
	row_count_t *xx = (row_count_t *)x;
	row_count_t *yy = (row_count_t *)y;
	return yy->count - xx->count;
}

static int compare_row_index(const void *x, const void *y) {
	row_count_t *xx = (row_count_t *)x;
	row_count_t *yy = (row_count_t *)y;
	return xx->row_index - yy->row_index;
}

static void reduce_matrix(msieve_obj *obj, uint32 *nrows, 
			uint32 dense_rows, uint32 *ncols, 
			la_col_t *cols) {

	/* Perform light filtering on the nrows x ncols
	   matrix specified by cols[]. The processing here is
	   limited to deleting columns that contain a singleton
	   row, deleting empty rows, and then deleting the heaviest
	   columns until the matrix has a few more columns than 
	   rows. Because deleting a column reduces the counts in 
	   several different rows, the process must iterate to 
	   convergence.
	   
	   We do *not* combine pairs of columns that contain a row
	   of weight 2. While this would reduce the size of the 
	   final system, the number of columns that can be combined
	   this way appears to be very small with effective NFS
	   filtering (something like 0.5% of the columns), and unlike
	   deleting singleton rows we cannot turn around and delete
	   heavy columns afterwards. It's not worth the effort.

	   Note that deleting singleton rows is not intended to 
	   make the Lanczos iteration run any faster (though it will); 
	   it's just that if we don't go to this trouble then the
	   matrix step could fail outright 
	   
	   Finally, note that the following completely ignores any
	   dense matrix rows that are stored implicitly. This means
	   NUM_EXTRA_RELATIONS has to be a little bigger than is
	   strictly necessary */

	uint32 r, c, i, j, k;
	uint32 passes;
	row_count_t *counts;
	uint32 reduced_rows;
	uint32 reduced_cols;

	/* sort the columns in order of increasing weight */

	qsort(cols, (size_t)(*ncols), sizeof(la_col_t), compare_weight);

	/* count the number of nonzero entries in each row */

	reduced_rows = *nrows;
	reduced_cols = *ncols;
	passes = 0;

	counts = (row_count_t *)malloc(reduced_rows * sizeof(row_count_t));
	for (i = 0; i < reduced_rows; i++) {
		counts[i].count = 0;
		counts[i].row_index = i;
	}
	for (i = 0; i < reduced_cols; i++) {
		for (j = 0; j < cols[i].weight; j++)
			counts[cols[i].data[j]].count++;
	}

	do {
		r = reduced_rows;

		/* remove any columns that contain the only entry
		   in one or more rows, then update the row counts
		   to reflect the missing column. Iterate until
		   no more columns can be deleted */

		do {
			c = reduced_cols;
			for (i = j = 0; i < reduced_cols; i++) {
				la_col_t *col = cols + i;
				for (k = 0; k < col->weight; k++) {
					if (counts[col->data[k]].count < 2)
						break;
				}
	
				if (k < col->weight) {
					for (k = 0; k < col->weight; k++) {
						counts[col->data[k]].count--;
					}
					free(col->data);
					free(col->cycle.list);
				}
				else {
					cols[j++] = cols[i];
				}
			}
			reduced_cols = j;
		} while (c != reduced_cols);
	
		/* count the number of rows that contain a
		   nonzero entry */

		for (i = reduced_rows = 0; i < *nrows; i++) {
			if (counts[i].count)
				reduced_rows++;
		}

		/* Because deleting a column reduces the weight
		   of many rows, the number of nonzero rows may
		   be much less than the number of columns. Delete
		   more columns until the matrix has the correct
		   aspect ratio. Columns at the end of cols[] are
		   the heaviest, so delete those (and update the
		   row counts again) */

		if (reduced_cols > reduced_rows + NUM_EXTRA_RELATIONS) {
			for (i = reduced_rows + NUM_EXTRA_RELATIONS;
					i < reduced_cols; i++) {

				la_col_t *col = cols + i;
				for (j = 0; j < col->weight; j++) {
					counts[col->data[j]].count--;
				}
				free(col->data);
				free(col->cycle.list);
			}
			reduced_cols = reduced_rows + NUM_EXTRA_RELATIONS;
		}

		/* if any columns were deleted in the previous step,
		   then the matrix is less dense and more columns
		   can be deleted; iterate until no further deletions
		   are possible */

		passes++;

	} while (r != reduced_rows);

	if (reduced_cols == 0) {
		free(counts);
		*nrows = reduced_rows;
		*ncols = reduced_cols;
		return;
	}

	/* later code will read in the cycles and dependencies
	   again from scratch, but we've rearranged the list
	   of matrix columns and have to remember the new order */

	dump_cycles(obj, cols, reduced_cols);

	if (dense_rows >= POST_LANCZOS_ROWS)
		dense_rows -= POST_LANCZOS_ROWS;
	else
		dense_rows = 0;

	logprintf(obj, "reduce to %u x %u in %u passes\n", 
			reduced_rows+dense_rows, reduced_cols, passes);

	/* permute the row indices to remove rows with zero
	   weight, put the heaviest row indices together,
	   and put each column in sorted order */
	
	qsort(counts, (size_t)(*nrows), 
			sizeof(row_count_t), compare_row_count);
	for (i = 0; i < *nrows; i++) {
		counts[i].count = i;
	}
	qsort(counts, (size_t)(*nrows), 
			sizeof(row_count_t), compare_row_index);

	for (i = 0; i < reduced_cols; i++) {
		la_col_t *col = cols + i;
		for (j = 0; j < col->weight; j++) {
			col->data[j] = counts[col->data[j]].count;
		}
		qsort(col->data, (size_t)col->weight, 
				sizeof(uint32), compare_uint32);
	}

	/* record the final matrix size */

	free(counts);
	*nrows = reduced_rows;
	*ncols = reduced_cols;
}

/*------------------------------------------------------------------*/
void nfs_solve_linear_system(msieve_obj *obj, 
			sieve_param_t *params, mp_t *n) {

	/* Generate linear dependencies */

	uint32 i, j;
	la_col_t *cols;
	uint32 nrows, ncols; 
	uint32 dense_rows, dense_row_words;
	uint32 total_weight;
	uint64 mask;
	uint64 *dependencies;
	uint64 *dense_submatrix = NULL;
	factor_base_t fb;

	logprintf(obj, "\n");
	logprintf(obj, "commencing linear algebra\n");

	/* read in the factor base */

	memset(&fb, 0, sizeof(fb));
	fb.rfb.max_prime = params->rfb_limit;
	fb.afb.max_prime = params->afb_limit;
	if (read_factor_base(obj, n, &fb)) {
		create_factor_base(obj, &fb, 1);
		write_factor_base(obj, n, &fb);
	}

	/* convert the list of relations from the sieving 
	   stage into a matrix. */

	total_weight = build_matrix(obj, &fb, &nrows, &dense_rows,
					&ncols, &cols);

	logprintf(obj, "matrix is %u x %u with weight %u (avg %5.2f/col)\n", 
				nrows + dense_rows, ncols, total_weight, 
				(double)total_weight / ncols);

	/* reduce the matrix dimensions to ignore nearly
	   or completely empty rows */

	reduce_matrix(obj, &nrows, dense_rows, &ncols, cols);

	if (ncols == 0) {
		logprintf(obj, "matrix is corrupt; skipping linear algebra\n");
		free(cols);
		return;
	}

	if (dense_rows) {
		/* save the first few dense rows for a separate
	   	   post-Lanczos linear algebra phase */

		mp_t tmp;
		dense_row_words = (dense_rows + 31) / 32;
		dense_submatrix = (uint64 *)malloc(ncols * sizeof(uint64));
		mask = (uint64)(-1) >> (64 - POST_LANCZOS_ROWS);
		mp_clear(&tmp);

		for (i = 0; i < ncols; i++) {
			uint32 *curr_row = cols[i].data + cols[i].weight;

			tmp.nwords = dense_row_words;
			for (j = 0; j < dense_row_words; j++)
				tmp.val[j] = curr_row[j];

			dense_submatrix[i] = ((uint64)tmp.val[0] |
				              (uint64)tmp.val[1] << 32) & mask;

			/* remove the first few dense rows from this
			   column of the matrix */

			if (dense_rows >= POST_LANCZOS_ROWS) {
				mp_rshift(&tmp, POST_LANCZOS_ROWS, &tmp);
				for (j = 0; j < dense_row_words; j++)
					curr_row[j] = tmp.val[j];
			}
		}
	}

	if (dense_rows >= POST_LANCZOS_ROWS)
		dense_rows -= POST_LANCZOS_ROWS;
	else
		dense_rows = 0;
	dense_row_words = (dense_rows + 31) / 32;
	nrows += dense_rows;
	if (dense_rows)
		logprintf(obj, "matrix includes %u packed rows\n", dense_rows);

	/* repack the matrix columns in the heap, now that 
	   they're the only major structure left in memory
	   and the number of words needed to store dense
	   rows has possibly been reduced. Also add dense_rows 
	   more rows to the matrix, by adding a bias to each 
	   row index */

	for (i = 0; i < ncols; i++) {
		la_col_t *c = cols + i;
		c->data = (uint32 *)realloc(c->data, (c->weight + 
					dense_row_words) * sizeof(uint32));
		if (dense_rows) {
			for (j = 0; j < c->weight; j++)
				c->data[j] += dense_rows;
		}
	}

	/* solve the linear system */

	do {
		dependencies = block_lanczos(obj, nrows, dense_rows,
						ncols, cols);
	} while(dependencies == NULL);

	for (i = 0; i < ncols; i++)
		free(cols[i].data);

	if (dense_submatrix) {
		combine_cols(ncols, dependencies, NULL, 
				dense_submatrix, NULL);
	}
	
	/* determine exactly how many of the dependencies
	   are trivial */

	for (i = 0, mask = 0; i < ncols; i++)
		mask |= dependencies[i];

	for (i = j = 0; i < 64; i++) {
		if (mask & ((uint64)(1) << i))
			j++;
	}
	if (j == 0) {
		logprintf(obj, "lanczos error: only trivial "
				"dependencies found\n");
	}
	else {
		logprintf(obj, "recovered %d nontrivial dependencies\n", j);
		dump_dependencies(obj, dependencies, ncols);
	}

	free(dependencies);
	free(dense_submatrix);
	free(cols);
}
