/*--------------------------------------------------------------------
This source distribution is placed in the public domain by its author,
Jason Papadopoulos. You may use it for any purpose, free of charge,
without having to notify anyone. I disclaim any responsibility for any
errors.

Optionally, please be nice and tell me if you find this source to be
useful. Again optionally, if you add to the functionality present here
please consider making those additions public too, so that others may 
benefit from your work.	
       				   --jasonp@boo.net 1/5/07
--------------------------------------------------------------------*/

#include "poly.h"
#include <gsl/gsl_errno.h>

/*------------------------------------------------------------------*/
static double integrand(double x, void *params) {

	/* callback for numerical integration. c[0] to
	   c[5] are polynomial coefficients, and c[6] is m */

	double *c = (double *)params;
	double val;

	val = c[0]+x*(c[1]+x*(c[2]+x*(c[3]+x*(c[4]+x*c[5]))));
	val = pow(fabs(val), -1.0/3.0) * pow(fabs(x - c[6]), -1.0/3.0);
	return val;
}

/*------------------------------------------------------------------*/
void analyze_poly(poly_config_t *config,
	      mp_t *m, mp_poly_t *poly) {

	/* analyze a polynomial for sieving goodness.
	   If the polynomial is good enough, save it */

	uint32 i, j;
	
	gsl_function F;
	double dcoeff[7];
	double size_score, size_error;
	double root_score;
	double combined_score;

	F.function = &integrand;
	F.params = dcoeff;

	dcoeff[0] = mp_signed_mp2d(&poly->coeff[0]);
	dcoeff[1] = mp_signed_mp2d(&poly->coeff[1]);
	dcoeff[2] = mp_signed_mp2d(&poly->coeff[2]);
	dcoeff[3] = mp_signed_mp2d(&poly->coeff[3]);
	dcoeff[4] = mp_signed_mp2d(&poly->coeff[4]);
	dcoeff[5] = mp_signed_mp2d(&poly->coeff[5]);
	dcoeff[6] = mp_mp2d(m);

	/* evaluate the polynomial at the predetermined
	   sample points */

	for (i = 0; i < NUM_POLY_SAMPLES; i++) {
		poly_sample_t *sample = config->samples + i;
		eval_poly(&sample->sample_val,
				sample->a, sample->b, poly);
	}

	/* Rather than use Murphy's method to rate the average
	   size of polynomial values, we use a result of D.J.
	   Bernstein: For rational poly R(x) and algebraic poly
	   A(x), the number of polynomial values where x is rational
	   and the size of A(x)*R(x) achieves a given value is 
	   proportional to the integral of

	   	dx / ((R(x) * A(x))^2) ^ (1/6)

	   for x from -infinity to +infinity. Larger values of
	   this integral imply more values that get found by
	   a siever.

	   We use a black box numerical integrator to compute
	   this integral. */
	   
	if (gsl_integration_qagi(&F, 0.0, 1.0e-6, (size_t)1000, 
				config->integrate_aux, 
				&size_score, &size_error) != 0) {
		return;
	}

	/* Now analyze the polynomial for root properties (i.e.
	   compute Murphy's 'alpha' value). For primes under 100
	   this uses the set of sample polynomial values, and
	   for larger primes we just count the roots that would
	   appear in a factor base */

	root_score = 0;
	for (i = 0; i < config->prime_list.num_primes; i++) {
		uint32 prime = config->prime_list.list[i];
		uint32 num_roots;

		if (prime < 100) {
			for (j = num_roots = 0; j < NUM_POLY_SAMPLES; j++) {
				signed_mp_t *sample_val = 
					    &config->samples[j].sample_val;
				while (mp_mod_1(&sample_val->num, 
						prime) == 0) {
					num_roots++;
					mp_divrem_1(&sample_val->num, prime,
							&sample_val->num);
				}
			}
			root_score += log((double)prime) * 
				(1.0 / (prime - 1) - 
				 (double)num_roots / NUM_POLY_SAMPLES);
		}
		else {
			uint32 dummy_roots[MAX_POLY_DEGREE];
			uint32 high_coeff;
			num_roots = poly_get_zeros(dummy_roots, poly, 
						prime, &high_coeff, 1);

			if (high_coeff == 0)
				num_roots++;

			root_score += (1.0 - (double)num_roots * prime / 
					(prime + 1)) * log((double)prime) / 
					(prime - 1);
		}
	}

	/* Another interesting problem: you have a root score and
	   a size score for each polynomial; how do you combine
	   them to get an overall rating? The value of root_score
	   implies that values of A(x) effectively have size
	   A(x) * exp(root_score). The second term is constant, so
	   plugging this expression into the integrand above gives: */

	combined_score = size_score * exp(-root_score/3);

	/* save the polynomial if its combined score is the 
	   best seen so far. A really thorough implementation
	   would do this in two stages: the best X polynomials
	   would all be kept in stage 1, and in stage 2 the 
	   complete list (possibly combined from the output of
	   different machines) would get a more accurate 
	   root_score computed for each polynomial. The top
	   few such polynomials would then be subjected to 
	   sieving experiments */

	if (combined_score > config->heap[0]->combined_score) {
		poly_select_t *entry = config->heap[0];

		entry->size_score = size_score;
		entry->root_score = root_score;
		entry->combined_score = combined_score;

		memcpy(&entry->apoly, poly, sizeof(mp_poly_t));

		entry->rpoly.degree = 1;
		entry->rpoly.coeff[1].sign = POSITIVE;
		entry->rpoly.coeff[1].num.nwords = 1;
		entry->rpoly.coeff[1].num.val[0] = 1;
		entry->rpoly.coeff[0].sign = NEGATIVE;
		mp_copy(m, &entry->rpoly.coeff[0].num);
	}
}

/*------------------------------------------------------------------*/
void poly_config_init(msieve_obj *obj, poly_config_t *config) {

	/* one-time initialization for polynomial search */

	uint32 i, num_primes;

	gsl_set_error_handler_off();

	fill_prime_list(&config->prime_list, 0xffff, PRIME_BOUND);
	num_primes = config->prime_list.num_primes;

	config->heap_num_filled = 0;
	for (i = 0; i < POLY_HEAP_SIZE; i++) {
		config->heap[i] = (poly_select_t *)calloc((size_t)1, 
					sizeof(poly_select_t));
	}

	config->integrate_aux = gsl_integration_workspace_alloc((size_t)1000);

	config->samples = (poly_sample_t *)calloc((size_t)NUM_POLY_SAMPLES,
						sizeof(poly_sample_t));

	/* when anayzing root properties, all polynomials
	   are sampled at the same set of points */

	i = 0;
	while (i < NUM_POLY_SAMPLES) {
		int32 a = get_rand(&obj->seed1, &obj->seed2);
		uint32 b = get_rand(&obj->seed1, &obj->seed2);

		if (mp_gcd_1((uint32)abs(a), b) == 1) {
			config->samples[i].a = a;
			config->samples[i].b = b;
			i++;
		}
	}
}

/*------------------------------------------------------------------*/
void poly_config_free(poly_config_t *config) {

	uint32 i;

	free(config->prime_list.list);
	gsl_integration_workspace_free(config->integrate_aux);
	for (i = 0; i < POLY_HEAP_SIZE; i++)
		free(config->heap[i]);
	free(config->samples);
}
