/*--------------------------------------------------------------------
This source distribution is placed in the public domain by its author,
Jason Papadopoulos. You may use it for any purpose, free of charge,
without having to notify anyone. I disclaim any responsibility for any
errors.

Optionally, please be nice and tell me if you find this source to be
useful. Again optionally, if you add to the functionality present here
please consider making those additions public too, so that others may 
benefit from your work.	
       				   --jasonp@boo.net 1/5/07
--------------------------------------------------------------------*/

#include "sqrt.h"

#define LOG2_IDEAL_HASHTABLE_SIZE 21

/*--------------------------------------------------------------------*/
static void eval_poly_derivative(mp_poly_t *poly, 
				mp_t *m1, signed_mp_t *m0, 
				mp_t *n, mp_t *res) {
	uint32 i;
	mp_t next_coeff;
	mp_t m1_pow, m0_tmp;

	mp_copy(m1, &m1_pow);
	mp_copy(&m0->num, &m0_tmp);
	if (m0->sign == POSITIVE)
		mp_sub(n, &m0_tmp, &m0_tmp);

	i = poly->degree;
	mp_mul_1(&poly->coeff[i].num, i, res);

	while (--i) {
		signed_mp_t *coeff = poly->coeff + i;

		mp_modmul(res, &m0_tmp, n, res);

		mp_mul_1(&coeff->num, i, &next_coeff);
		if (coeff->sign == NEGATIVE)
			mp_sub(n, &next_coeff, &next_coeff);
		mp_modmul(&next_coeff, &m1_pow, n, &next_coeff);
		mp_add(res, &next_coeff, res);

		if (i > 1)
			mp_modmul(&m1_pow, m1, n, &m1_pow);
	}

	if (mp_cmp(res, n) >= 0)
		mp_sub(res, n, res);
}

/*--------------------------------------------------------------------*/
static uint32 rat_square_root(relation_t *rlist, uint32 num_relations,
				mp_t *n, mp_t *sqrt_r) {
	uint32 i, j;
	hashtable_t h;
	uint32 already_seen;
	mp_t base, exponent, tmp;

	hashtable_init(&h, LOG2_IDEAL_HASHTABLE_SIZE, 10000, 1);
	for (i = 0; i < num_relations; i++) {
		relation_t *r = rlist + i;
		for (j = 0; j < r->num_factors_r; j++) {
			hash_t *curr_ideal = hashtable_find(&h,
						r->factors_r + j, 
						&already_seen);
			if (!already_seen)
				curr_ideal->payload[1] = r->refcnt;
			else
				curr_ideal->payload[1] += r->refcnt;
		}
	}

	mp_clear(&base); base.nwords = 1;
	mp_clear(&exponent); exponent.nwords = 1;
	mp_clear(sqrt_r); sqrt_r->nwords = sqrt_r->val[0] = 1;

	for (i = 1; i < h.match_array_size; i++) {
		hash_t *curr_ideal = h.match_array + i;
		if (curr_ideal->payload[1] % 2) {
			hashtable_free(&h);
			return 1;
		}
		if (curr_ideal->payload[0]) {
			base.val[0] = curr_ideal->payload[0];
			exponent.val[0] = curr_ideal->payload[1] / 2;
			mp_expo(&base, &exponent, n, &tmp);
			mp_modmul(sqrt_r, &tmp, n, sqrt_r);
		}
	}

	hashtable_free(&h);
	return 0;
}

/*--------------------------------------------------------------------*/
static uint32 verify_alg_ideal_powers(relation_t *rlist, 
					uint32 num_relations) {
	uint32 i, j, k;
	hashtable_t h;
	uint32 already_seen;
	uint32 full_num_relations = 0;
	uint32 *counts = NULL;
	uint32 count_alloc = 0;

	hashtable_init(&h, LOG2_IDEAL_HASHTABLE_SIZE, 10000, 2);

	for (i = 0; i < num_relations; i++) {
		relation_t *r = rlist + i;
		relation_lp_t rlp;

		find_large_ideals(r, &rlp, 0, 0);

		for (j = 0; j < rlp.ideal_count; j++) {
			ideal_t *curr_ideal = rlp.ideal_list + j;
			hash_t *curr_ideal_hash;

			if (curr_ideal->i.rat_or_alg == RATIONAL_IDEAL)
				continue;

			curr_ideal_hash = hashtable_find(&h,
					curr_ideal, &already_seen);
			k = curr_ideal_hash - h.match_array;

			if (!already_seen && 
			    count_alloc < h.match_array_alloc) {
				counts = (uint32 *)realloc(counts,
						h.match_array_alloc *
						sizeof(uint32));
				memset(counts + count_alloc, 0,
						(h.match_array_alloc -
				       		count_alloc) * sizeof(uint32));
				count_alloc = h.match_array_alloc;
			}
			counts[k] += r->refcnt;
		}
		full_num_relations += r->refcnt;
	}

	for (i = 1; i < h.match_array_size; i++) {
		if (counts[i] % 2) {
			full_num_relations = 0;
			break;
		}
	}

	hashtable_free(&h);
	free(counts);
	return full_num_relations;
}

/*--------------------------------------------------------------------*/
void nfs_find_factors(msieve_obj *obj, mp_t *n) {

	uint32 i, j;
	factor_base_t fb;
	mp_poly_t monic_alg_poly;
	mp_t sqrt_r, sqrt_a;
	mp_t c, tmp1, tmp2;
	signed_mp_t *m0;
	mp_t *m1; 
	factor_list_t factor_list;
	uint32 dep_lower = 1;
	uint32 dep_upper = 64;

	logprintf(obj, "\n");
	logprintf(obj, "commencing square root phase\n");

	/* read in the factor base */

	memset(&fb, 0, sizeof(fb));
	if (read_poly(obj, n, &fb.rfb.poly, &fb.afb.poly)) {
		printf("polynomials not found\n");
		exit(-1);
	}

	if (fb.rfb.poly.degree != 1) {
		printf("cannot handle non-linear polynomials\n");
		exit(-1);
	}
	m0 = &fb.rfb.poly.coeff[0];
	m1 = &fb.rfb.poly.coeff[1].num;

	/* construct a monic version of the algebraic poly */

	j = fb.afb.poly.degree;
	memset(&monic_alg_poly, 0, sizeof(mp_poly_t));
	mp_copy(&fb.afb.poly.coeff[j].num, &c);
	mp_copy(&c, &tmp1);
	monic_alg_poly.coeff[j-1] = fb.afb.poly.coeff[j-1];
	for (i = j - 2; (int32)i >= 0; i--) {
		mp_mul(&fb.afb.poly.coeff[i].num, &tmp1,
				&monic_alg_poly.coeff[i].num);
		monic_alg_poly.coeff[i].sign = fb.afb.poly.coeff[i].sign;

		if (i > 0) {
			mp_mul(&c, &tmp1, &tmp2);
			mp_copy(&tmp2, &tmp1);
		}
	}
	monic_alg_poly.coeff[j].num.nwords = 1;
	monic_alg_poly.coeff[j].num.val[0] = 1;
	monic_alg_poly.degree = fb.afb.poly.degree;

	factor_list_init(&factor_list, n, 1);
	if (obj->nfs_lower && obj->nfs_upper) {
		dep_lower = MIN(obj->nfs_lower, 64);
		dep_upper = MIN(obj->nfs_upper, 64);
		dep_upper = MAX(dep_lower, dep_upper);
	}

	for (i = dep_lower; i <= dep_upper; i++) {
		uint32 k, m;
		uint32 num_relations;
		uint32 full_num_relations;
		relation_t *rlist;
		abpair_t *abpairs;
		uint32 num_cycles;
		la_col_t *cycle_list;
		mp_t exponent;

		logprintf(obj, "reading relations for dependency %u\n", i);

		nfs_read_cycles(obj, &fb, &num_cycles, &cycle_list, 
				&num_relations, &rlist, 0, i);

		if (num_cycles == 0)
			continue;

		free_cycle_list(cycle_list, num_cycles);

		full_num_relations = verify_alg_ideal_powers(
					rlist, num_relations);
		if (full_num_relations == 0) {
			logprintf(obj, "algebraic side is not a square!\n");
			nfs_free_relation_list(rlist, num_relations);
			continue;
		}
		if (full_num_relations % 2) {
			logprintf(obj, "number of relations is not even\n");
			nfs_free_relation_list(rlist, num_relations);
			continue;
		}
		if (rat_square_root(rlist, num_relations, n, &sqrt_r) != 0) {
			logprintf(obj, "rational side is not a square!\n");
			nfs_free_relation_list(rlist, num_relations);
			continue;
		}

		abpairs = (abpair_t *)malloc(full_num_relations *
						sizeof(abpair_t));
		for (j = k = 0; j < num_relations; j++) {
			relation_t *r = rlist + j;
			for (m = 0; m < r->refcnt; m++) {
				abpairs[k].a = r->a;
				abpairs[k].b = r->b;
				k++;
			}
		}
		nfs_free_relation_list(rlist, num_relations);

		mp_clear(&sqrt_a);
		alg_square_root(obj, &monic_alg_poly, n, &c, m1, m0, abpairs, 
					full_num_relations, &sqrt_a);
		free(abpairs);
		if (mp_is_zero(&sqrt_a)) {
			logprintf(obj, "algebraic square root failed\n");
			continue;
		}

		mp_clear(&exponent);
		exponent.nwords = 1;
		exponent.val[0] = full_num_relations / 2 + 
					fb.afb.poly.degree - 2;
		mp_expo(&c, &exponent, n, &tmp1);
		eval_poly_derivative(&fb.afb.poly, m1, m0, n, &tmp2);
		mp_modmul(&tmp1, &tmp2, n, &tmp1);
		mp_modmul(&tmp1, &sqrt_r, n, &sqrt_r);

		if (!mp_is_one(m1)) {
			exponent.val[0] = full_num_relations / 2;
			mp_expo(m1, &exponent, n, &tmp1);
			mp_modmul(&tmp1, &sqrt_a, n, &sqrt_a);
		}

		mp_modmul(&sqrt_r, &sqrt_r, n, &tmp1);
		mp_modmul(&sqrt_a, &sqrt_a, n, &tmp2);
		if (mp_cmp(&tmp1, &tmp2) != 0) {
			logprintf(obj, "dependency does not form a "
					"congruence of squares!\n");
			continue;
		}

		mp_add(&sqrt_r, &sqrt_a, &tmp1);
		mp_gcd(&tmp1, n, &tmp1);
		if (!mp_is_one(&tmp1) && mp_cmp(n, &tmp1) != 0) {
			/* nontrivial factor found; add it to the
			   list of factors. Note that we quit if any
			   remaining composite factors total less than
			   about 70 digits; in that case, it's faster
			   to run MPQS than to wait for another dependency
			   to finish */
			if (factor_list_add(obj, &factor_list, &tmp1) < 230)
				break;
		}
	}

	factor_list_free(&factor_list, obj);
}
