/*--------------------------------------------------------------------
This source distribution is placed in the public domain by its author,
Jason Papadopoulos. You may use it for any purpose, free of charge,
without having to notify anyone. I disclaim any responsibility for any
errors.

Optionally, please be nice and tell me if you find this source to be
useful. Again optionally, if you add to the functionality present here
please consider making those additions public too, so that others may 
benefit from your work.	
       				   --jasonp@boo.net 1/5/07
--------------------------------------------------------------------*/

#include "sqrt.h"

#define NUM_POLY_COEFFS (MAX_POLY_DEGREE + 1)

typedef struct {
	uint32 degree;
	ap_t coeff[NUM_POLY_COEFFS];
} ap_poly_t;

typedef struct {
	ap_poly_t *monic_poly;
	abpair_t *rlist;
	mp_t *c;
	fastmult_info_t *mult_info;
} relation_prod_t;

/*-------------------------------------------------------------------*/
static void ap_poly_init(ap_poly_t *poly) {
	
	uint32 i;

	for (i = poly->degree = 0; i < NUM_POLY_COEFFS; i++)
		ap_init(poly->coeff + i);
}

/*-------------------------------------------------------------------*/
static void ap_poly_clear(ap_poly_t *poly) {
	
	uint32 i;

	for (i = poly->degree = 0; i < NUM_POLY_COEFFS; i++)
		ap_clear(poly->coeff + i);
}

/*-------------------------------------------------------------------*/
static void ap_poly_mod_q(ap_poly_t *p, ap_t *q, ap_poly_t *res,
			ap_t *recip, fastmult_info_t *info) {

	uint32 i;

	for (i = 0; i <= p->degree; i++)
		ap_mod(p->coeff + i, q, recip, res->coeff + i, info);

	while (--i) {
		if (!ap_is_zero(&res->coeff[i]))
			break;
	}
	res->degree = i;
}

/*-------------------------------------------------------------------*/
static void ap_poly_bits(ap_poly_t *p, uint32 *total_bits, uint32 *max_bits) {

	uint32 i, bits1, bits2;

	for (i = bits1 = bits2 = 0; i <= p->degree; i++) {
		uint32 curr_bits = ap_bits(p->coeff + i);
		bits1 += curr_bits;
		bits2 = MAX(bits2, curr_bits);
	}

	*total_bits = bits1;
	*max_bits = bits2;
}

/*-------------------------------------------------------------------*/
static void ap_poly_monic_derivative(ap_poly_t *src, ap_poly_t *dest,
					fastmult_info_t *info) {
	
	uint32 i;
	ap_t c;

	ap_init(&c);
	for (i = 0; i < src->degree; i++) {
		ap_si2ap(i + 1, POSITIVE, &c);
		ap_mul(src->coeff + (i+1), &c, dest->coeff + i, info);
	}
	ap_si2ap(i + 1, POSITIVE, dest->coeff + i);

	dest->degree = src->degree;
	for (i++; i < NUM_POLY_COEFFS; i++)
		ap_clear(dest->coeff + i);
	ap_clear(&c);
}

/*-------------------------------------------------------------------*/
static void ap_poly_mul(ap_poly_t *p1, ap_poly_t *p2,
			ap_poly_t *mod, fastmult_info_t *info,
			uint32 free_p2) {

	uint32 i, j;
	uint32 d = mod->degree;
	uint32 d1 = p1->degree;
	uint32 d2 = p2->degree;
	uint32 prod_degree;
	ap_t tmp[NUM_POLY_COEFFS + 1];
	ap_t prod;

	ap_init(&prod);
	for (i = 0; i < NUM_POLY_COEFFS + 1; i++) {
		ap_init(tmp + i);
	}

	for (i = 0; i <= d1; i++) {
		ap_mul(p1->coeff + i, p2->coeff + d2, tmp + i, info);
	}
	prod_degree = d1;
	if (free_p2) {
		ap_clear(p2->coeff + d2);
	}

	for (i = d2 - 1; (int32)i >= 0; i--) {

		ap_t high_coeff = tmp[prod_degree + 1];
		for (j = prod_degree; (int32)j >= 0; j--) {
			tmp[j + 1] = tmp[j];
		}
		tmp[0] = high_coeff;
		tmp[0].nwords = 0;

		for (j = d1; j; j--) {
			ap_mul(p1->coeff + j, p2->coeff + i, &prod, info);
			ap_add(tmp + j, &prod, tmp + j);
		}
		ap_mul(p1->coeff + j, p2->coeff + i, tmp + j, info);
		if (free_p2) {
			ap_clear(p2->coeff + i);
		}

		prod_degree = d + 1;
		while (prod_degree && ap_is_zero(tmp + prod_degree))
			prod_degree--;

		if (prod_degree <= d)
			continue;

		for (j = d; (int32)j >= 0; j--) {
			ap_mul(mod->coeff + j, tmp + prod_degree, &prod, info);
			ap_sub(tmp + j, &prod, tmp + j);
		}
		prod_degree--;
	}

	ap_clear(&prod);
	for (i = 0; i <= prod_degree; i++) {
		ap_clear(p1->coeff + i);
		p1->coeff[i] = tmp[i];
	}
	for (; i < NUM_POLY_COEFFS + 1; i++)
		ap_clear(tmp + i);

	i = prod_degree;
	while (i > 0 && ap_is_zero(p1->coeff + i)) {
		mp_clear(p1->coeff + i);
		i--;
	}
	p1->degree = i;
}

/*-------------------------------------------------------------------*/
static void relation_to_poly(abpair_t *abpair, mp_t *c,
				ap_poly_t *poly) {

	mp_t tmp;

	mp_mul_1(c, (uint32)abs(abpair->a), &tmp);
	ap_mp2ap(&tmp, (uint32)(abpair->a >= 0 ? POSITIVE : NEGATIVE),
			poly->coeff + 0);
	ap_si2ap(abpair->b, NEGATIVE, poly->coeff + 1);
	poly->degree = 1;
	if (ap_is_zero(poly->coeff + 1))
		poly->degree = 0;
}

/*-------------------------------------------------------------------*/
static void multiply_relations(relation_prod_t *prodinfo, 
			uint32 index1, uint32 index2,
			ap_poly_t *prod) {

	uint32 i;
	ap_poly_t prod1, prod2;

	if (index1 == index2) {
		relation_to_poly(prodinfo->rlist + index1,
				 prodinfo->c, prod);
		return;
	}

	ap_poly_init(&prod1);
	ap_poly_init(&prod2);
	
	if (index1 == index2 - 1) {
		relation_to_poly(prodinfo->rlist + index1,
				 prodinfo->c, &prod1);
		relation_to_poly(prodinfo->rlist + index2,
				 prodinfo->c, &prod2);
	}
	else {
		uint32 mid = (index1 + index2) / 2;
		multiply_relations(prodinfo, index1, mid, &prod1);
		multiply_relations(prodinfo, mid + 1, index2, &prod2);
	}

	ap_poly_mul(&prod1, &prod2, prodinfo->monic_poly, 
			prodinfo->mult_info, 1);

	prod->degree = prod1.degree;
	for (i = 0; i <= prod1.degree; i++)
		prod->coeff[i] = prod1.coeff[i];
}

/*-------------------------------------------------------------------*/
#define NUM_PRIME_RETRIES 1000

static void get_initial_inv_sqrt(msieve_obj *obj, mp_poly_t *mp_alg_poly,
				ap_poly_t *prod, ap_poly_t *isqrt_mod_q, 
				ap_t *q_out) {

	uint32 i, j;
	uint32 q;
	prime_sieve_t prime_sieve;
	ap_poly_t prod_mod_q;
	mp_poly_t mp_prod_mod_q;
	mp_poly_t mp_isqrt_mod_q;

	init_prime_sieve(&prime_sieve, q_out->val[0], (uint32)(-1));
	ap_poly_init(&prod_mod_q);

	for (i = 0; i < NUM_PRIME_RETRIES; i++) {
		do {
			q = get_next_prime(&prime_sieve);
		} while (!is_irreducible(mp_alg_poly, q));

		ap_si2ap(q, POSITIVE, q_out);
		ap_poly_mod_q(prod, q_out, &prod_mod_q, NULL, NULL);

		memset(&mp_prod_mod_q, 0, sizeof(mp_poly_t));
		mp_prod_mod_q.degree = prod_mod_q.degree;
		for (j = 0; j <= prod_mod_q.degree; j++) {
			signed_mp_t *coeff = mp_prod_mod_q.coeff + j;
			coeff->sign = prod_mod_q.coeff[j].sign;
			coeff->num.nwords = 1;
			coeff->num.val[0] = prod_mod_q.coeff[j].val[0];
		}

		if (inv_sqrt_mod_q(&mp_isqrt_mod_q, &mp_prod_mod_q, 
				mp_alg_poly, q, &obj->seed1, &obj->seed2)) {
			break;
		}
	}

	free_prime_sieve(&prime_sieve);
	ap_poly_clear(&prod_mod_q);
	if (i == NUM_PRIME_RETRIES) {
		printf("error: cannot recover square root mod q\n");
		exit(-1);
	}

	for (i = 0; i <= mp_isqrt_mod_q.degree; i++) {
		signed_mp_t *coeff = mp_isqrt_mod_q.coeff + i;
		ap_mp2ap(&coeff->num, coeff->sign, isqrt_mod_q->coeff + i);
	}
	isqrt_mod_q->degree = mp_isqrt_mod_q.degree;

	logprintf(obj, "initial square root is modulo %u\n", q);
}

/*-------------------------------------------------------------------*/
static uint32 get_final_sqrt(msieve_obj *obj, ap_poly_t *alg_poly,
			ap_poly_t *prod, ap_poly_t *isqrt_mod_q, 
			ap_t *q, fastmult_info_t *info) {

	uint32 i, j;
	uint32 qbits;
	uint32 rbits;
	uint32 prod_bits, prod_max_bits;
	uint32 num_iter;
	ap_poly_t tmp_poly;
	ap_t three;
	ap_t recip;
	uint32 converged = 0;

	ap_poly_init(&tmp_poly);
	ap_init(&three);
	ap_init(&recip);
	ap_si2ap(3, POSITIVE, &three);
	ap_poly_bits(prod, &prod_bits, &prod_max_bits);

	i = q->val[0];
	for (num_iter = 0; ap_bits(q) < prod_max_bits / 2 + 4000; num_iter++)
		ap_mul(q, q, q, info);

	ap_recip(q, &recip, prod_max_bits - ap_bits(q), info);
	ap_poly_mod_q(prod, q, prod, &recip, info);
	ap_si2ap(i, POSITIVE, q);
	ap_clear(&recip);
	for (i = 0; i <= prod->degree; i++) {
		ap_t *coeff = prod->coeff + i;
		coeff->val = (uint32 *)realloc(coeff->val, 
					coeff->nwords * sizeof(uint32));
		coeff->num_alloc = coeff->nwords;
	}

	for (i = 0; i < num_iter; i++) {

		ap_mul(q, q, q, info);
		qbits = ap_bits(q);

		if (obj->flags & (MSIEVE_FLAG_USE_LOGFILE |
	    			   MSIEVE_FLAG_LOG_TO_STDOUT)) {
			fprintf(stderr, "precision = %u bits\r", qbits);
			fflush(stderr);
		}

		if (q->nwords > MAX_MP_WORDS) {
			rbits = ap_bits(prod->coeff + 0);
			if (qbits >= rbits)
				rbits = 20000000;
			else
				rbits = MIN(rbits - qbits, 20000000);

			ap_recip(q, &recip, rbits, info);
		}

		ap_poly_mod_q(prod, q, &tmp_poly, &recip, info);
		ap_poly_mul(&tmp_poly, isqrt_mod_q, alg_poly, info, 0);
		ap_poly_mod_q(&tmp_poly, q, &tmp_poly, &recip, info);
		ap_poly_mul(&tmp_poly, isqrt_mod_q, alg_poly, info, 0);
		ap_poly_mod_q(&tmp_poly, q, &tmp_poly, &recip, info);

		ap_sub(&tmp_poly.coeff[0], &three, &tmp_poly.coeff[0]);
		for (j = 0; j <= tmp_poly.degree; j++) {
			ap_t *coeff = tmp_poly.coeff + j;

			if (coeff->sign == POSITIVE)
				coeff->sign = NEGATIVE;
			else
				coeff->sign = POSITIVE;

			if (!ap_is_zero(coeff) && coeff->val[0] % 2)
				ap_add(coeff, q, coeff);
			ap_rshift(coeff, 1, coeff);
		}

		ap_poly_mul(&tmp_poly, isqrt_mod_q, alg_poly, info, 1);
		ap_poly_mod_q(&tmp_poly, q, isqrt_mod_q, &recip, info);
		ap_poly_clear(&tmp_poly);

		if (i == num_iter - 1) {
			ap_poly_mul(isqrt_mod_q, prod, alg_poly, info, 1);
			ap_poly_mod_q(isqrt_mod_q, q, isqrt_mod_q, 
							&recip, info);

			for (j = 0; j <= isqrt_mod_q->degree; j++) {
				ap_t *coeff = isqrt_mod_q->coeff + j;
				if (q->nwords == coeff->nwords &&
				    !memcmp(q->val + q->nwords - 4,
					    coeff->val + coeff->nwords - 4,
					    4 * sizeof(uint32))) {
					if (coeff->sign == NEGATIVE)
						ap_add(coeff, q, coeff);
					else
						ap_sub(coeff, q, coeff);
				}
			}
			ap_poly_bits(isqrt_mod_q, &prod_bits, &j);
			if (prod_bits < (isqrt_mod_q->degree + 1) * qbits - 100)
				converged = 1;
		}

		ap_clear(&recip);
	}

	if (obj->flags & (MSIEVE_FLAG_USE_LOGFILE |
	    	   MSIEVE_FLAG_LOG_TO_STDOUT))
		printf("\n");

	if (!converged)
		logprintf(obj, "Newton iteration failed to converge\n");

	ap_poly_clear(&tmp_poly);
	ap_clear(&three);
	ap_clear(&recip);
	return converged;
}

/*-------------------------------------------------------------------*/
static void convert_to_integer(ap_poly_t *alg_sqrt, mp_t *n,
				mp_t *c, mp_t *m1, signed_mp_t *m0, 
				mp_t *res) {
	uint32 i;
	ap_t ap_n;
	ap_t *ap_coeff;
	mp_t m1_pow;
	mp_t m0_tmp;
	mp_t next_coeff;

	ap_init(&ap_n); 
	ap_mp2ap(n, POSITIVE, &ap_n);
	ap_poly_mod_q(alg_sqrt, &ap_n, alg_sqrt, NULL, NULL);
	ap_clear(&ap_n);

	mp_copy(m1, &m1_pow);
	mp_modmul(&m0->num, c, n, &m0_tmp);
	if (m0->sign == POSITIVE)
		mp_sub(n, &m0_tmp, &m0_tmp);

	i = alg_sqrt->degree;
	ap_coeff = alg_sqrt->coeff + i;
	res->nwords = ap_coeff->nwords;
	memcpy(res->val, ap_coeff->val, res->nwords * sizeof(uint32));
	if (ap_coeff->sign == NEGATIVE)
		mp_sub(n, res, res);

	for (i--; (int32)i >= 0; i--) {
		mp_modmul(res, &m0_tmp, n, res);

		ap_coeff = alg_sqrt->coeff + i;
		mp_clear(&next_coeff);
		next_coeff.nwords = ap_coeff->nwords;
		memcpy(next_coeff.val, ap_coeff->val, 
				next_coeff.nwords * sizeof(uint32));
		mp_modmul(&next_coeff, &m1_pow, n, &next_coeff);
		if (ap_coeff->sign == NEGATIVE)
			mp_sub(n, &next_coeff, &next_coeff);

		mp_add(res, &next_coeff, res);
		if (i > 0)
			mp_modmul(&m1_pow, m1, n, &m1_pow);
	}
	if (mp_cmp(res, n) > 0)
		mp_sub(res, n, res);
}

/*-------------------------------------------------------------------*/
void alg_square_root(msieve_obj *obj, mp_poly_t *mp_alg_poly, 
			mp_t *n, mp_t *c, mp_t *m1, signed_mp_t *m0,
			abpair_t *rlist, uint32 num_relations,
			mp_t *sqrt_a) {
	
	uint32 i;
	ap_poly_t alg_poly;
	ap_poly_t d_alg_poly;
	ap_poly_t prod;
	ap_poly_t alg_sqrt;
	fastmult_info_t info;
	relation_prod_t prodinfo;
	double log2_prodsize;
	ap_t *p;
	ap_t q;

	ap_init(&q);
	ap_poly_init(&alg_poly);
	ap_poly_init(&d_alg_poly);
	ap_poly_init(&prod);
	ap_poly_init(&alg_sqrt);
	fastmult_info_init(&info);

	for (i = 0; i < mp_alg_poly->degree; i++) {
		signed_mp_t *coeff = mp_alg_poly->coeff + i;
		ap_mp2ap(&coeff->num, coeff->sign, alg_poly.coeff + i);
	}
	alg_poly.degree = mp_alg_poly->degree - 1;

	prodinfo.monic_poly = &alg_poly;
	prodinfo.rlist = rlist;
	prodinfo.c = c;
	prodinfo.mult_info = &info;

	logprintf(obj, "multiplying %u relations\n", num_relations);
	multiply_relations(&prodinfo, 0, num_relations - 1, &prod);
	logprintf(obj, "multiply complete, coefficients have about "
			"%3.2lf million bits\n",
			(double)ap_bits(prod.coeff + 0) / 1e6);

	ap_poly_monic_derivative(&alg_poly, &d_alg_poly, &info);
	ap_poly_mul(&d_alg_poly, &d_alg_poly, &alg_poly, &info, 0);
	ap_poly_mul(&prod, &d_alg_poly, &alg_poly, &info, 1);

	p = &prod.coeff[0];
	log2_prodsize = 32.0 * (p->nwords - 2) +
		        log(p->val[p->nwords - 1] * MP_RADIX +
			    p->val[p->nwords - 2]) / M_LN2 + 10000;
	while (log2_prodsize > 31.5)
		log2_prodsize *= 0.5;

	ap_si2ap((uint32)pow(2.0, log2_prodsize) + 1, POSITIVE, &q);
	get_initial_inv_sqrt(obj, mp_alg_poly, &prod, &alg_sqrt, &q);
	if (get_final_sqrt(obj, &alg_poly, &prod, &alg_sqrt, &q, &info))
		convert_to_integer(&alg_sqrt, n, c, m1, m0, sqrt_a);

	ap_poly_clear(&prod);
	ap_poly_clear(&alg_sqrt);
	ap_poly_clear(&alg_poly);
	ap_poly_clear(&d_alg_poly);
	ap_clear(&q);
	fastmult_info_free(&info);
}
