/*--------------------------------------------------------------------
This source distribution is placed in the public domain by its author,
Jason Papadopoulos. You may use it for any purpose, free of charge,
without having to notify anyone. I disclaim any responsibility for any
errors.

Optionally, please be nice and tell me if you find this source to be
useful. Again optionally, if you add to the functionality present here
please consider making those additions public too, so that others may 
benefit from your work.	
       				   --jasonp@boo.net 1/5/07
--------------------------------------------------------------------*/

#include <common.h>
#include "mpqs.h"

static uint32 build_matrix(uint32 ncols, la_col_t *cols, 
		    	relation_t *relation_list,
			uint32 start_cols);

static void reduce_matrix(msieve_obj *obj,
			uint32 *nrows, 
			uint32 *ncols, 
			la_col_t *cols);

/*------------------------------------------------------------------*/
void solve_linear_system(msieve_obj *obj, uint32 fb_size, 
		    uint64 **bitfield, relation_t *relation_list, 
		    la_col_t *cycle_list, uint32 *num_cycles) {

	/* Generate linear dependencies among the relations
	   in full_relations and partial_relations */

	uint32 i, j;
	la_col_t *cols;
	uint32 nrows, ncols;
	uint32 total_weight;
	uint64 mask;
	uint64 *dependencies;

	ncols = MIN(*num_cycles, fb_size + NUM_EXTRA_RELATIONS);
	nrows = fb_size;
	cols = cycle_list;

	/* convert the list of relations from the sieving 
	   stage into a matrix. */

	total_weight = build_matrix(ncols, cols, relation_list, *num_cycles);

	logprintf(obj, "%u x %u system, weight %u (avg %4.2f/col)\n", 
				fb_size, ncols, total_weight, 
				(double)total_weight / ncols);

	/* reduce the matrix dimensions to ignore almost
	   empty rows */

	reduce_matrix(obj, &nrows, &ncols, cols);

	if (ncols == 0) {
		logprintf(obj, "matrix is corrupt; skipping linear algebra\n");
		free(cols);
		*num_cycles = 0;
		return;
	}

	/* solve the linear system */

	do {
		dependencies = block_lanczos(obj, nrows, 0, ncols, cols);
	} while(dependencies == NULL);

	for (i = 0; i < ncols; i++)
		free(cols[i].data);

	/* determine exactly how many of the dependencies
	   are trivial */

	for (i = 0, mask = 0; i < ncols; i++)
		mask |= dependencies[i];

	for (i = j = 0; i < 64; i++) {
		if (mask & ((uint64)(1) << i))
			j++;
	}
	if (j == 0) {
		logprintf(obj, "lanczos error: only trivial "
				"dependencies found\n");
		free(dependencies);
		return;
	}

	logprintf(obj, "recovered %d nontrivial dependencies\n", j);
	*bitfield = dependencies;
	*num_cycles = ncols;
}

/*------------------------------------------------------------------*/
void reduce_matrix(msieve_obj *obj, uint32 *nrows, 
			uint32 *ncols, la_col_t *cols) {

	/* Perform light filtering on the nrows x ncols
	   matrix specified by cols[]. The processing here is
	   limited to deleting columns that contain a singleton
	   row, then resizing the matrix to have a few more
	   columns than rows. Because deleting a column reduces
	   the counts in several different rows, the process
	   must iterate to convergence.
	   
	   Note that this step is not intended to make the Lanczos
	   iteration run any faster (though it will); it's just
	   that if we don't go to this trouble then there are 
	   factorizations for which the matrix step will fail 
	   outright  */

	uint32 r, c, i, j, k;
	uint32 passes;
	uint32 *counts;
	uint32 reduced_rows;
	uint32 reduced_cols;

	/* count the number of nonzero entries in each row */

	counts = (uint32 *)calloc((size_t)*nrows, sizeof(uint32));
	for (i = 0; i < *ncols; i++) {
		for (j = 0; j < cols[i].weight; j++)
			counts[cols[i].data[j]]++;
	}

	reduced_rows = *nrows;
	reduced_cols = *ncols;
	passes = 0;

	do {
		r = reduced_rows;

		/* remove any columns that contain the only entry
		   in one or more rows, then update the row counts
		   to reflect the missing column. Iterate until
		   no more columns can be deleted */

		do {
			c = reduced_cols;
			for (i = j = 0; i < reduced_cols; i++) {
				la_col_t *col = cols + i;
				for (k = 0; k < col->weight; k++) {
					if (counts[col->data[k]] < 2)
						break;
				}
	
				if (k < col->weight) {
					for (k = 0; k < col->weight; k++) {
						counts[col->data[k]]--;
					}
					free(col->data);
					free(col->cycle.list);
				}
				else {
					cols[j++] = cols[i];
				}
			}
			reduced_cols = j;
		} while (c != reduced_cols);
	
		/* count the number of rows that contain a
		   nonzero entry */

		for (i = reduced_rows = 0; i < *nrows; i++) {
			if (counts[i])
				reduced_rows++;
		}

		/* Because deleting a column reduces the weight
		   of many rows, the number of nonzero rows may
		   be much less than the number of columns. Delete
		   more columns until the matrix has the correct
		   aspect ratio. Columns at the end of cols[] are
		   the heaviest, so delete those (and update the
		   row counts again) */

		if (reduced_cols > reduced_rows + NUM_EXTRA_RELATIONS) {
			for (i = reduced_rows + NUM_EXTRA_RELATIONS;
					i < reduced_cols; i++) {

				la_col_t *col = cols + i;
				for (j = 0; j < col->weight; j++) {
					counts[col->data[j]]--;
				}
				free(col->data);
				free(col->cycle.list);
			}
			reduced_cols = reduced_rows + NUM_EXTRA_RELATIONS;
		}

		/* if any columns were deleted in the previous step,
		   then the matrix is less dense and more columns
		   can be deleted; iterate until no further deletions
		   are possible */

		passes++;

	} while (r != reduced_rows);

	logprintf(obj, "reduce to %u x %u in %u passes\n", 
			reduced_rows, reduced_cols, passes);

	free(counts);

	/* record the final matrix size. Note that we can't touch
	   nrows because all the column data (and the sieving relations
	   that produced it) would have to be updated */

	*ncols = reduced_cols;
}

/*------------------------------------------------------------------*/
static uint32 merge_relations(uint32 *merge_array,
		  uint32 *src1, uint32 n1,
		  uint32 *src2, uint32 n2) {

	/* Given two sorted lists of integers, merge
	   the lists into a single sorted list with
	   duplicate entries removed. If a particular
	   entry occurs an even number of times in the
	   two input lists, don't add it to the final list
	   at all. Returns the number of elements in the
	   resulting list */

	uint32 i1, i2, val1, val2, count1, count2;
	uint32 num_merge;

	i1 = i2 = 0;
	num_merge = 0;

	while (i1 < n1 && i2 < n2) {
		val1 = src1[i1];
		val2 = src2[i2];

		if (val1 < val2) {
			count1 = 0;
			do {
				i1++; count1++;
			} while (i1 < n1 && src1[i1] == val1);

			if (count1 & 1)
				merge_array[num_merge++] = val1;
		}
		else if (val1 > val2) {
			count2 = 0;
			do {
				i2++; count2++;
			} while (i2 < n2 && src2[i2] == val2);

			if (count2 & 1)
				merge_array[num_merge++] = val2;
		}
		else {
			count1 = count2 = 0;
			do {
				i1++; count1++;
			} while (i1 < n1 && src1[i1] == val1);
			do {
				i2++; count2++;
			} while (i2 < n2 && src2[i2] == val2);

			if ( (count1 + count2) & 1 )
				merge_array[num_merge++] = val1;
		}
	}

	if (i2 == n2) {
		src2 = src1;
		i2 = i1;
		n2 = n1;
	}

	while (i2 < n2) {
		count2 = 0; val2 = src2[i2];

		do {
			i2++; count2++;
		} while (i2 < n2 && src2[i2] == val2);

		if (count2 & 1)
			merge_array[num_merge++] = val2;
	}

	return num_merge;
}

/*--------------------------------------------------------------------*/
static int compare_weight(const void *x, const void *y) {
	la_col_t *xx = (la_col_t *)x;
	la_col_t *yy = (la_col_t *)y;
	return xx->weight - yy->weight;
}

/*------------------------------------------------------------------*/
#define MAX_COL_WEIGHT 1000

static uint32 build_matrix(uint32 ncols, la_col_t *cols, 
			   relation_t *relation_list,
			   uint32 start_cols) {

	/* Convert lists of relations from the sieving stage
	   into a sparse matrix. The matrix is stored by
	   columns, pointed to by 'cols'. The total number 
	   of nonzero entries in the matrix is returned */

	uint32 i, j;
	uint32 total_weight = 0;
	la_col_t *col;

	/* Cycles are assumed to be sorted in order of increasing
	   number of relations, so that any cycles that are
	   not used would have created the heaviest matrix columns
	   anyway */

	for (i = 0; i < ncols; i++) {
		uint32 buf[MAX_COL_WEIGHT];
		uint32 accum[MAX_COL_WEIGHT];
		uint32 weight;

		/* merge each succeeding relation into the accumulated
		   matrix column */

		col = cols + i;

		for (j = weight = 0; j < col->cycle.num_relations; j++) {
			relation_t *r = relation_list + col->cycle.list[j];
			weight = merge_relations(accum, buf, weight,
						r->fb_offsets, r->num_factors);
			memcpy(buf, accum, weight * sizeof(uint32));
		}

		total_weight += weight;
		col->weight = weight;
		col->data = (uint32 *)malloc(weight * sizeof(uint32));
		memcpy(col->data, buf, weight * sizeof(uint32));
	}

	for (; i < start_cols; i++) {
		free(cols[i].cycle.list);
		cols[i].cycle.list = NULL;
	}

	/* sort the final matrix to put the heaviest columns
	   at the end. The previous sort order put the cycles
	   with the most factors last, but that only approximates
	   the heaviest matrix columns */

	qsort(cols, (size_t)ncols, sizeof(la_col_t), compare_weight);

	return total_weight;
}
